"""
Common utility functions for main storage compaction tests.

These functions are shared between incremental and multiple rounds compaction tests.
"""
# Allow classes to use self-referencing Type hints in Python 3.7.
from __future__ import annotations
from enum import Enum
from typing import Any, Dict, List, Optional, Tuple
import datetime as dt
from datetime import timezone

import tempfile
import os
import shutil

import pyarrow as pa


from deltacat.tests.compute.test_util_constant import (
    BASE_TEST_SOURCE_NAMESPACE,
    BASE_TEST_SOURCE_TABLE_NAME,
    BASE_TEST_SOURCE_TABLE_VERSION,
    BASE_TEST_DESTINATION_NAMESPACE,
    BASE_TEST_DESTINATION_TABLE_NAME,
    BASE_TEST_DESTINATION_TABLE_VERSION,
    REBASING_NAMESPACE,
    REBASING_TABLE_NAME,
    REBASING_TABLE_VERSION,
)
from deltacat.compute.compactor import (
    RoundCompletionInfo,
)
from deltacat.compute.compactor.model.compaction_session_audit_info import (
    CompactionSessionAuditInfo,
)
from deltacat.storage.model.partition import (
    PartitionLocator,
    PartitionScheme,
    PartitionKey as StoragePartitionKey,
)
from deltacat.storage.model.stream import StreamLocator
from deltacat.storage.model.table_version import TableVersionLocator
from deltacat.storage.model.table import TableLocator
from deltacat.storage.model.namespace import NamespaceLocator
from deltacat.storage.model.sort_key import (
    SortScheme,
)
from deltacat.storage.model.delta import (
    Delta,
    DeltaType,
)
from deltacat.storage.model.partition import (
    Partition,
    PartitionKeyList,
)
from deltacat.storage.model.stream import Stream
from deltacat.storage.model.transform import IdentityTransform
from deltacat.storage.model.schema import Schema
from deltacat.compute.compactor.model.compactor_version import CompactorVersion

from deltacat.storage import metastore
from deltacat.catalog.model.properties import CatalogProperties


class PartitionKeyType(str, Enum):
    INT = "int"
    STRING = "string"
    TIMESTAMP = "timestamp"


class PartitionKey(dict):
    @staticmethod
    def of(key_name: str, key_type: PartitionKeyType) -> PartitionKey:
        return PartitionKey({"keyName": key_name, "keyType": key_type.value})

    @property
    def key_name(self) -> str:
        return self["keyName"]

    @property
    def key_type(self) -> PartitionKeyType:
        key_type = self["keyType"]
        return None if key_type is None else PartitionKeyType(key_type)


"""
UTILS
"""


def get_test_partition_locator(partition_id):
    tv_locator = TableVersionLocator.of(
        TableLocator.of(NamespaceLocator.of("default"), "test_table"), "1"
    )
    stream_locator = StreamLocator.of(tv_locator, "test_stream_id", "local")
    partition_locator = PartitionLocator.of(
        stream_locator, partition_id=partition_id, partition_values=[]
    )

    return partition_locator


def create_main_deltacat_storage_kwargs() -> Dict[str, Any]:
    """
    Helper function to create main deltacat storage kwargs

    Returns: kwargs to use for main deltacat storage, i.e. {"catalog": CatalogProperties(...)}
    """
    temp_dir = tempfile.mkdtemp()
    catalog = CatalogProperties(root=temp_dir)
    return {"catalog": catalog}


def clean_up_main_deltacat_storage_kwargs(storage_kwargs: Dict[str, Any]):
    """
    Cleans up directory created by create_main_deltacat_storage_kwargs
    """
    catalog = storage_kwargs["catalog"]
    if hasattr(catalog, "root") and os.path.exists(catalog.root):
        shutil.rmtree(catalog.root)


def _create_table_main(
    namespace: str,
    table_name: str,
    table_version: str,
    sort_keys: Optional[List[Any]],
    partition_keys: Optional[List[PartitionKey]],
    input_deltas: Optional[pa.Table],
    ds_mock_kwargs: Optional[Dict[str, Any]],
):
    """
    Main storage version of _create_table that works for both incremental and multiple rounds tests.

    For incremental tests, input_deltas is provided to extract schema.
    For multiple rounds tests, input_deltas can be None and we use a simpler approach.
    """
    # Create namespace first
    metastore.create_namespace(namespace=namespace, **ds_mock_kwargs)

    # Handle schema creation
    if input_deltas is not None:
        # Incremental test approach - extract schema from input deltas
        schema = input_deltas.schema

        # Add partition key fields to schema if they're not already present
        if partition_keys:
            for pk in partition_keys:
                field_name = pk.key_name
                if field_name not in schema.names:
                    # Add partition key field with appropriate type
                    if pk.key_type == PartitionKeyType.INT:
                        field_type = pa.int32()
                    elif pk.key_type == PartitionKeyType.STRING:
                        field_type = pa.string()
                    elif (
                        pk.key_type == PartitionKeyType.TIMESTAMP
                    ):  # Handle timestamp type properly
                        field_type = pa.timestamp("us")
                    else:
                        field_type = pa.string()  # Default to string

                    schema = schema.append(pa.field(field_name, field_type))

        schema_obj = Schema.of(schema=schema)
    else:
        # Multiple rounds test approach - use None for schema (will be set later)
        schema_obj = None

    sort_scheme = SortScheme.of(sort_keys) if sort_keys else None

    # Convert test partition keys to storage partition keys
    storage_partition_keys = []
    if partition_keys:
        for pk in partition_keys:
            storage_partition_key = StoragePartitionKey.of(
                key=[pk.key_name],
                name=pk.key_name,
                transform=IdentityTransform.of(),
            )
            storage_partition_keys.append(storage_partition_key)

    # Create partition scheme
    partition_scheme = None
    if storage_partition_keys:
        partition_scheme = PartitionScheme.of(
            keys=PartitionKeyList.of(storage_partition_keys),
            scheme_id="default_partition_scheme",
        )

    # Create table version (which creates table and stream automatically)
    metastore.create_table_version(
        namespace=namespace,
        table_name=table_name,
        table_version=table_version,
        schema=schema_obj,
        partition_scheme=partition_scheme,
        sort_keys=sort_scheme,
        **ds_mock_kwargs,
    )

    return namespace, table_name, table_version


def create_src_table_main(
    sort_keys: Optional[List[Any]],
    partition_keys: Optional[List[PartitionKey]],
    input_deltas: Optional[pa.Table],
    ds_mock_kwargs: Optional[Dict[str, Any]],
):
    """
    Main storage version of create_src_table
    """
    source_namespace: str = BASE_TEST_SOURCE_NAMESPACE
    source_table_name: str = BASE_TEST_SOURCE_TABLE_NAME
    source_table_version: str = BASE_TEST_SOURCE_TABLE_VERSION
    return _create_table_main(
        source_namespace,
        source_table_name,
        source_table_version,
        sort_keys,
        partition_keys,
        input_deltas,
        ds_mock_kwargs,
    )


def create_destination_table_main(
    sort_keys: Optional[List[Any]],
    partition_keys: Optional[List[PartitionKey]],
    input_deltas: Optional[pa.Table],
    ds_mock_kwargs: Optional[Dict[str, Any]],
):
    """
    Main storage version of create_destination_table
    """
    destination_namespace: str = BASE_TEST_DESTINATION_NAMESPACE
    destination_table_name: str = BASE_TEST_DESTINATION_TABLE_NAME
    destination_table_version: str = BASE_TEST_DESTINATION_TABLE_VERSION
    return _create_table_main(
        destination_namespace,
        destination_table_name,
        destination_table_version,
        sort_keys,
        partition_keys,
        input_deltas,
        ds_mock_kwargs,
    )


def create_rebase_table_main(
    sort_keys: Optional[List[Any]],
    partition_keys: Optional[List[PartitionKey]],
    input_deltas: Optional[pa.Table],
    ds_mock_kwargs: Optional[Dict[str, Any]],
):
    """
    Main storage version of create_rebase_table
    """
    rebasing_namespace = REBASING_NAMESPACE
    rebasing_table_name = REBASING_TABLE_NAME
    rebasing_table_version = REBASING_TABLE_VERSION
    return _create_table_main(
        rebasing_namespace,
        rebasing_table_name,
        rebasing_table_version,
        sort_keys,
        partition_keys,
        input_deltas,
        ds_mock_kwargs,
    )


def get_rci_from_partition(
    partition_locator: PartitionLocator, deltacat_storage=None, **kwargs
) -> RoundCompletionInfo:
    """
    Read RoundCompletionInfo from a partition metafile.

    Args:
        partition_locator: Locator of the partition containing the RoundCompletionInfo
        deltacat_storage: Storage implementation (defaults to metastore)
        **kwargs: Additional arguments to pass to deltacat_storage.get_partition (e.g., catalog)

    Returns:
        RoundCompletionInfo object from the partition, or None if not found
    """
    from deltacat.storage import metastore

    if deltacat_storage is None:
        deltacat_storage = metastore

    partition = deltacat_storage.get_partition(
        partition_locator.stream_locator, partition_locator.partition_values, **kwargs
    )

    if partition and partition.compaction_round_completion_info:
        return partition.compaction_round_completion_info

    return None


def _add_deltas_to_partition_main(
    deltas_ingredients: List[Tuple[pa.Table, DeltaType, Optional[Dict[str, str]]]],
    partition: Optional[Partition],
    ds_mock_kwargs: Optional[Dict[str, Any]],
) -> Tuple[Optional[Delta], int]:
    """
    Add deltas to a partition using main storage
    """
    all_deltas_length = 0
    incremental_delta = None
    for (delta_data, delta_type, delete_parameters) in deltas_ingredients:
        staged_delta: Delta = metastore.stage_delta(
            delta_data,
            partition,
            delta_type,
            entry_params=delete_parameters,
            **ds_mock_kwargs,
        )
        incremental_delta = metastore.commit_delta(
            staged_delta,
            **ds_mock_kwargs,
        )
        all_deltas_length += len(delta_data) if delta_data else 0
    return incremental_delta, all_deltas_length


def add_late_deltas_to_partition_main(
    late_deltas: List[Tuple[pa.Table, DeltaType, Optional[Dict[str, str]]]],
    source_partition: Optional[Partition],
    ds_mock_kwargs: Optional[Dict[str, Any]],
) -> Tuple[Optional[Delta], int]:
    """
    Add late deltas to a partition using main storage
    """
    return _add_deltas_to_partition_main(late_deltas, source_partition, ds_mock_kwargs)


def multiple_rounds_create_src_w_deltas_destination_rebase_w_deltas_strategy_main(
    sort_keys: Optional[List[Any]],
    partition_keys: Optional[List[PartitionKey]],
    input_deltas: List[pa.Table],
    partition_values: Optional[List[Any]],
    ds_mock_kwargs: Optional[Dict[str, Any]],
) -> Tuple[Stream, Stream, Optional[Stream], bool]:
    """
    Main storage version of multiple_rounds_create_src_w_deltas_destination_rebase_w_deltas_strategy
    """
    # For multiple rounds, we need to extract the first delta to get schema
    first_delta_table = input_deltas[0][0] if input_deltas else None
    source_namespace, source_table_name, source_table_version = create_src_table_main(
        sort_keys, partition_keys, first_delta_table, ds_mock_kwargs
    )

    source_table_stream: Stream = metastore.get_stream(
        namespace=source_namespace,
        table_name=source_table_name,
        table_version=source_table_version,
        **ds_mock_kwargs,
    )

    # Convert partition values to correct types
    converted_partition_values = []
    if partition_values and partition_keys:
        for i, (value, pk) in enumerate(zip(partition_values, partition_keys)):
            if pk.key_type == PartitionKeyType.INT:
                converted_partition_values.append(int(value))
            else:
                converted_partition_values.append(value)
    else:
        converted_partition_values = partition_values

    staged_partition: Partition = metastore.stage_partition(
        source_table_stream,
        converted_partition_values,
        partition_scheme_id="default_partition_scheme" if partition_keys else None,
        **ds_mock_kwargs,
    )

    is_delete = False
    input_delta_length = 0
    for (
        input_delta,
        input_delta_type,
        input_delta_parameters,
    ) in input_deltas:
        if input_delta_type is DeltaType.DELETE:
            is_delete = True
        staged_delta = metastore.stage_delta(
            input_delta,
            staged_partition,
            input_delta_type,
            entry_params=input_delta_parameters,
            **ds_mock_kwargs,
        )
        metastore.commit_delta(staged_delta, **ds_mock_kwargs)
        input_delta_length += len(input_delta)
    metastore.commit_partition(staged_partition, **ds_mock_kwargs)

    (
        destination_table_namespace,
        destination_table_name,
        destination_table_version,
    ) = create_destination_table_main(
        sort_keys, partition_keys, first_delta_table, ds_mock_kwargs
    )
    destination_table_stream: Stream = metastore.get_stream(
        namespace=destination_table_namespace,
        table_name=destination_table_name,
        table_version=destination_table_version,
        **ds_mock_kwargs,
    )

    # Always create rebase table for multiple rounds tests
    (
        rebasing_table_namespace,
        rebasing_table_name,
        rebasing_table_version,
    ) = create_rebase_table_main(
        sort_keys, partition_keys, first_delta_table, ds_mock_kwargs
    )
    rebasing_table_stream: Stream = metastore.get_stream(
        namespace=rebasing_table_namespace,
        table_name=rebasing_table_name,
        table_version=rebasing_table_version,
        **ds_mock_kwargs,
    )

    # Stage partition and add deltas to rebase table
    rebased_staged_partition: Partition = metastore.stage_partition(
        rebasing_table_stream,
        converted_partition_values,
        partition_scheme_id="default_partition_scheme" if partition_keys else None,
        **ds_mock_kwargs,
    )

    for (
        input_delta,
        input_delta_type,
        input_delta_parameters,
    ) in input_deltas:
        staged_delta = metastore.stage_delta(
            input_delta,
            rebased_staged_partition,
            input_delta_type,
            entry_params=input_delta_parameters,
            **ds_mock_kwargs,
        )
        metastore.commit_delta(staged_delta, **ds_mock_kwargs)
    metastore.commit_partition(rebased_staged_partition, **ds_mock_kwargs)

    return (
        source_table_stream,
        destination_table_stream,
        rebasing_table_stream,
        is_delete,
    )


def create_src_w_deltas_destination_plus_destination_main(
    sort_keys: Optional[List[Any]],
    partition_keys: Optional[List[PartitionKey]],
    input_deltas: pa.Table,
    input_delta_type: DeltaType,
    partition_values: Optional[List[Any]],
    ds_mock_kwargs: Optional[Dict[str, Any]],
    simulate_is_inplace: bool = False,
) -> Tuple[Stream, Stream, Optional[Stream], str, str, str]:
    """
    Create source with deltas and destination tables for incremental compaction testing
    """
    source_namespace, source_table_name, source_table_version = create_src_table_main(
        sort_keys, partition_keys, input_deltas, ds_mock_kwargs
    )

    source_table_stream: Stream = metastore.get_stream(
        namespace=source_namespace,
        table_name=source_table_name,
        table_version=source_table_version,
        **ds_mock_kwargs,
    )

    # Convert partition values to correct types
    converted_partition_values = []
    if partition_values and partition_keys:
        for i, (value, pk) in enumerate(zip(partition_values, partition_keys)):
            if pk.key_type == PartitionKeyType.INT:
                converted_partition_values.append(int(value))
            else:
                converted_partition_values.append(value)
    else:
        converted_partition_values = partition_values

    staged_partition: Partition = metastore.stage_partition(
        source_table_stream,
        converted_partition_values,
        partition_scheme_id="default_partition_scheme" if partition_keys else None,
        **ds_mock_kwargs,
    )
    metastore.commit_delta(
        metastore.stage_delta(
            input_deltas, staged_partition, input_delta_type, **ds_mock_kwargs
        ),
        **ds_mock_kwargs,
    )
    metastore.commit_partition(staged_partition, **ds_mock_kwargs)
    source_table_stream_after_committed: Stream = metastore.get_stream(
        namespace=source_namespace,
        table_name=source_table_name,
        table_version=source_table_version,
        **ds_mock_kwargs,
    )

    destination_table_namespace: Optional[str] = None
    destination_table_name: Optional[str] = None
    destination_table_version: Optional[str] = None
    if not simulate_is_inplace:
        (
            destination_table_namespace,
            destination_table_name,
            destination_table_version,
        ) = create_destination_table_main(
            sort_keys, partition_keys, input_deltas, ds_mock_kwargs
        )
    else:
        destination_table_namespace = source_namespace
        destination_table_name = source_table_name
        destination_table_version = source_table_version

    destination_table_stream: Stream = metastore.get_stream(
        namespace=destination_table_namespace,
        table_name=destination_table_name,
        table_version=destination_table_version,
        **ds_mock_kwargs,
    )

    return (
        source_table_stream_after_committed,
        destination_table_stream,
        None,
        source_namespace,
        source_table_name,
        source_table_version,
    )


def create_src_w_deltas_destination_rebase_w_deltas_strategy_main(
    sort_keys: Optional[List[Any]],
    partition_keys: Optional[List[PartitionKey]],
    input_deltas: pa.Table,
    input_delta_type: DeltaType,
    partition_values: Optional[List[Any]],
    ds_mock_kwargs: Optional[Dict[str, Any]],
) -> Tuple[Stream, Stream, Optional[Stream]]:
    """
    Main storage version of create_src_w_deltas_destination_rebase_w_deltas_strategy

    Creates source table with deltas, destination table, and rebase table for rebase testing.
    This test scenario sets up different source and rebase partition locators to simulate
    scenarios like hash bucket count changes.
    """
    from deltacat.utils.common import current_time_ms

    last_stream_position = current_time_ms()
    source_namespace, source_table_name, source_table_version = create_src_table_main(
        sort_keys, partition_keys, input_deltas, ds_mock_kwargs
    )

    source_table_stream: Stream = metastore.get_stream(
        namespace=source_namespace,
        table_name=source_table_name,
        table_version=source_table_version,
        **ds_mock_kwargs,
    )

    # Convert partition values to correct types, including timestamp handling
    converted_partition_values = []
    if partition_values and partition_keys:
        for i, (value, pk) in enumerate(zip(partition_values, partition_keys)):
            if pk.key_type == PartitionKeyType.INT:
                converted_partition_values.append(int(value))
            elif pk.key_type == PartitionKeyType.TIMESTAMP:
                # Handle timestamp partition values
                if isinstance(value, str) and "T" in value and value.endswith("Z"):
                    import pandas as pd

                    ts = pd.to_datetime(value)
                    # Convert to microseconds since epoch for PyArrow timestamp[us]
                    converted_partition_values.append(int(ts.timestamp() * 1_000_000))
                else:
                    converted_partition_values.append(value)
            else:
                converted_partition_values.append(value)
    else:
        converted_partition_values = partition_values

    staged_partition: Partition = metastore.stage_partition(
        source_table_stream,
        converted_partition_values,
        partition_scheme_id="default_partition_scheme" if partition_keys else None,
        **ds_mock_kwargs,
    )
    staged_delta: Delta = metastore.stage_delta(
        input_deltas, staged_partition, input_delta_type, **ds_mock_kwargs
    )
    staged_delta.locator.stream_position = last_stream_position
    metastore.commit_delta(staged_delta, **ds_mock_kwargs)
    metastore.commit_partition(staged_partition, **ds_mock_kwargs)

    source_table_stream_after_committed: Stream = metastore.get_stream(
        namespace=source_namespace,
        table_name=source_table_name,
        table_version=source_table_version,
        **ds_mock_kwargs,
    )

    # Create the destination table
    (
        destination_table_namespace,
        destination_table_name,
        destination_table_version,
    ) = create_destination_table_main(
        sort_keys, partition_keys, input_deltas, ds_mock_kwargs
    )

    # Create the rebase table
    (
        rebase_table_namespace,
        rebase_table_name,
        rebase_table_version,
    ) = create_rebase_table_main(
        sort_keys, partition_keys, input_deltas, ds_mock_kwargs
    )

    rebasing_table_stream: Stream = metastore.get_stream(
        namespace=rebase_table_namespace,
        table_name=rebase_table_name,
        table_version=rebase_table_version,
        **ds_mock_kwargs,
    )

    staged_partition: Partition = metastore.stage_partition(
        rebasing_table_stream,
        converted_partition_values,
        partition_scheme_id="default_partition_scheme" if partition_keys else None,
        **ds_mock_kwargs,
    )
    staged_delta: Delta = metastore.stage_delta(
        input_deltas, staged_partition, **ds_mock_kwargs
    )
    staged_delta.locator.stream_position = last_stream_position
    metastore.commit_delta(staged_delta, **ds_mock_kwargs)
    metastore.commit_partition(staged_partition, **ds_mock_kwargs)

    # Get destination stream
    destination_table_stream: Stream = metastore.get_stream(
        namespace=destination_table_namespace,
        table_name=destination_table_name,
        table_version=destination_table_version,
        **ds_mock_kwargs,
    )

    rebased_stream_after_committed: Stream = metastore.get_stream(
        namespace=rebase_table_namespace,
        table_name=rebase_table_name,
        table_version=rebase_table_version,
        **ds_mock_kwargs,
    )

    return (
        source_table_stream_after_committed,
        destination_table_stream,
        rebased_stream_after_committed,
    )


def create_incremental_deltas_on_source_table_main(
    source_namespace: str,
    source_table_name: str,
    source_table_version: str,
    source_table_stream: Stream,
    partition_values_param,
    incremental_deltas: List[Tuple[pa.Table, DeltaType, Optional[Dict[str, str]]]],
    ds_mock_kwargs: Optional[Dict[str, Any]] = None,
) -> Tuple[PartitionLocator, Delta, int, bool]:
    """
    Main storage version of create_incremental_deltas_on_source_table
    """
    total_records = 0
    has_delete_deltas = False
    new_delta = None

    # Convert partition values for partition lookup (same as in other helper functions)
    converted_partition_values_for_lookup = partition_values_param
    if (
        partition_values_param
        and source_table_stream.partition_scheme
        and source_table_stream.partition_scheme.keys
    ):
        converted_partition_values_for_lookup = []

        # Get partition field names from the storage partition scheme
        storage_partition_keys = source_table_stream.partition_scheme.keys
        partition_field_names = []

        for storage_key in storage_partition_keys:
            # Each storage PartitionKey has a 'key' property that contains FieldLocators
            # Extract the field name from the first FieldLocator
            field_name = storage_key.key[0] if storage_key.key else None
            partition_field_names.append(field_name)

        for i, value in enumerate(partition_values_param):
            # For timestamp fields like 'region_id', we need to convert the timestamp string
            if i < len(partition_field_names):
                field_name = partition_field_names[i]

                # Check if this is likely a timestamp field based on the value format
                if isinstance(value, str) and "T" in value and value.endswith("Z"):
                    # This looks like a timestamp string - convert it
                    import pandas as pd

                    ts = pd.to_datetime(value)
                    # Convert to microseconds since epoch for PyArrow timestamp[us]
                    converted_partition_values_for_lookup.append(
                        int(ts.timestamp() * 1_000_000)
                    )
                elif isinstance(value, str) and value.isdigit():
                    # This looks like an integer string
                    converted_partition_values_for_lookup.append(int(value))
                else:
                    # Keep as-is
                    converted_partition_values_for_lookup.append(value)
            else:
                converted_partition_values_for_lookup.append(value)

    # Get the current partition to stage deltas against
    try:
        source_partition: Partition = metastore.get_partition(
            source_table_stream.locator,
            converted_partition_values_for_lookup,
            **ds_mock_kwargs,
        )
    except Exception:
        # If we can't get the partition, it might not exist yet. Try to create it.
        # Stage a new partition if it doesn't exist
        staged_partition: Partition = metastore.stage_partition(
            source_table_stream,
            converted_partition_values_for_lookup,
            partition_scheme_id="default_partition_scheme"
            if source_table_stream.partition_scheme
            else None,
            **ds_mock_kwargs,
        )
        # Commit the empty partition first
        metastore.commit_partition(staged_partition, **ds_mock_kwargs)

        # Now try to get it again
        source_partition: Partition = metastore.get_partition(
            source_table_stream.locator,
            converted_partition_values_for_lookup,
            **ds_mock_kwargs,
        )

    if source_partition is None:
        raise ValueError(
            f"Could not create or retrieve partition for values: {converted_partition_values_for_lookup}"
        )

    for delta_table, delta_type, properties_dict in incremental_deltas:
        # Skip None deltas (empty incremental deltas)
        if delta_table is None:
            continue

        total_records += len(delta_table)

        if delta_type == DeltaType.DELETE:
            has_delete_deltas = True

        # Stage and commit the delta
        staged_delta: Delta = metastore.stage_delta(
            delta_table,
            source_partition,
            delta_type,
            entry_params=properties_dict,
            **ds_mock_kwargs,
        )
        new_delta = metastore.commit_delta(staged_delta, **ds_mock_kwargs)

    # If all deltas were None, return None for new_delta
    if new_delta is None:
        return None, None, total_records, has_delete_deltas

    # Get updated stream after deltas were committed
    source_table_stream_after_committed: Stream = metastore.get_stream(
        source_namespace,
        source_table_name,
        source_table_version,
        **ds_mock_kwargs,
    )

    # Get updated partition after deltas were committed
    source_partition_after_committed: Partition = metastore.get_partition(
        source_table_stream_after_committed.locator,
        converted_partition_values_for_lookup,
        **ds_mock_kwargs,
    )

    return (
        source_partition_after_committed.locator,
        new_delta,
        total_records,
        has_delete_deltas,
    )


def get_compacted_delta_locator_from_partition(
    partition_locator: PartitionLocator, deltacat_storage=None, **kwargs
):
    """
    Get compacted delta locator from partition RoundCompletionInfo.

    Args:
        partition_locator: Locator of the partition containing the RoundCompletionInfo
        deltacat_storage: Storage implementation (defaults to metastore)
        **kwargs: Additional arguments to pass to get_rci_from_partition (e.g., catalog)

    Returns:
        DeltaLocator of the compacted delta
    """
    round_completion_info: RoundCompletionInfo = get_rci_from_partition(
        partition_locator, deltacat_storage, **kwargs
    )

    if round_completion_info:
        return round_completion_info.compacted_delta_locator
    return None


def offer_iso8601_timestamp_list(
    periods: int,
    unit_of_time: str,
    end_time=dt.datetime(2023, 5, 3, 10, 0, 0, 0, tzinfo=timezone.utc),
) -> List[str]:
    """
    Returns a list of ISO 8601 timestamps, each periods units of time before the start time.

    Args:
    periods: The number of timestamps to return.
    unit_of_time: The unit of time to use for the timestamps. Must be one of "seconds", "minutes", "hours", "days", or "weeks".
    end_time: The end time for the timestamps. Defaults to 2023-05-03T10:00:00Z.

    Returns:
    A list of ISO 8601 timestamps, each periods units of time before the start time.

    Raises:
    ValueError: If the unit_of_time argument is not one of "seconds", "minutes", "hours", "days", or "weeks".
    """
    import datetime as dt

    UTC_ISO_8601_FORMAT_WITHOUT_MILLIS = "%Y-%m-%dT%H:%M:%SZ"  # '2018-09-05T14:09:03Z'

    acceptable_units_of_time = ["seconds", "minutes", "hours", "days", "weeks"]
    if unit_of_time not in acceptable_units_of_time:
        raise ValueError(
            f"unit_of_time {unit_of_time} is not supported. Please use one of these time units: {acceptable_units_of_time}"
        )
    res = []
    for i in range(periods):
        kwarg = {unit_of_time: i}
        res.append(
            (end_time - dt.timedelta(**kwarg)).strftime(
                UTC_ISO_8601_FORMAT_WITHOUT_MILLIS
            )
        )
    return res


def assert_compaction_audit(
    compactor_version: CompactorVersion, compaction_audit: CompactionSessionAuditInfo
):
    if compactor_version == CompactorVersion.V2:
        audit_entries = [
            compaction_audit.deltacat_version,
            compaction_audit.ray_version,
            compaction_audit.audit_url,
            compaction_audit.hash_bucket_count,
            compaction_audit.compactor_version,
            compaction_audit.input_size_bytes,
            compaction_audit.input_file_count,
            compaction_audit.estimated_in_memory_size_bytes_during_discovery,
            compaction_audit.uniform_deltas_created,
            compaction_audit.delta_discovery_time_in_seconds,
            compaction_audit.hash_bucket_time_in_seconds,
            compaction_audit.hash_bucket_invoke_time_in_seconds,
            compaction_audit.hash_bucket_result_size_bytes,
            compaction_audit.total_cluster_object_store_memory_bytes,
            compaction_audit.total_cluster_memory_bytes,
            compaction_audit.total_object_store_memory_used_bytes,
            compaction_audit.hash_bucket_post_object_store_memory_used_bytes,
            compaction_audit.hash_bucket_result_wait_time_in_seconds,
            compaction_audit.peak_memory_used_bytes_per_hash_bucket_task,
            compaction_audit.hash_bucket_processed_size_bytes,
            compaction_audit.input_records,
            compaction_audit.merge_time_in_seconds,
            compaction_audit.merge_invoke_time_in_seconds,
            compaction_audit.merge_result_size,
            compaction_audit.merge_post_object_store_memory_used_bytes,
            compaction_audit.merge_result_wait_time_in_seconds,
            compaction_audit.peak_memory_used_bytes_per_merge_task,
            compaction_audit.records_deduped,
            compaction_audit.records_deleted,
            compaction_audit.compaction_time_in_seconds,
            compaction_audit.peak_memory_used_bytes_by_compaction_session_process,
            compaction_audit.untouched_file_count,
            compaction_audit.untouched_file_ratio,
            compaction_audit.untouched_record_count,
            compaction_audit.untouched_size_bytes,
            compaction_audit.output_file_count,
            compaction_audit.output_size_bytes,
            compaction_audit.output_size_pyarrow_bytes,
            compaction_audit.peak_memory_used_bytes_per_task,
            compaction_audit.pyarrow_version,
            compaction_audit.telemetry_time_in_seconds,
            compaction_audit.observed_input_inflation,
            compaction_audit.observed_input_average_record_size_bytes,
        ]
        for entry in audit_entries:
            assert entry is not None
    return True


def assert_compaction_audit_no_hash_bucket(
    compactor_version: CompactorVersion, compaction_audit: CompactionSessionAuditInfo
):
    if compactor_version == CompactorVersion.V2:
        audit_entries = [
            compaction_audit.deltacat_version,
            compaction_audit.ray_version,
            compaction_audit.audit_url,
            compaction_audit.hash_bucket_count,
            compaction_audit.compactor_version,
            compaction_audit.input_size_bytes,
            compaction_audit.input_file_count,
            compaction_audit.estimated_in_memory_size_bytes_during_discovery,
            compaction_audit.uniform_deltas_created,
            compaction_audit.delta_discovery_time_in_seconds,
            compaction_audit.total_cluster_object_store_memory_bytes,
            compaction_audit.total_cluster_memory_bytes,
            compaction_audit.total_object_store_memory_used_bytes,
            compaction_audit.input_records,
            compaction_audit.merge_time_in_seconds,
            compaction_audit.merge_invoke_time_in_seconds,
            compaction_audit.merge_result_size,
            compaction_audit.merge_post_object_store_memory_used_bytes,
            compaction_audit.merge_result_wait_time_in_seconds,
            compaction_audit.peak_memory_used_bytes_per_merge_task,
            compaction_audit.records_deduped,
            compaction_audit.records_deleted,
            compaction_audit.compaction_time_in_seconds,
            compaction_audit.peak_memory_used_bytes_by_compaction_session_process,
            compaction_audit.untouched_file_count,
            compaction_audit.untouched_file_ratio,
            compaction_audit.untouched_record_count,
            compaction_audit.untouched_size_bytes,
            compaction_audit.output_file_count,
            compaction_audit.output_size_bytes,
            compaction_audit.output_size_pyarrow_bytes,
            compaction_audit.peak_memory_used_bytes_per_task,
            compaction_audit.pyarrow_version,
            compaction_audit.telemetry_time_in_seconds,
        ]
        for entry in audit_entries:
            assert entry is not None
    return True


def read_audit_file(audit_file_path: str, catalog_root: str) -> Dict[str, Any]:
    """
    Read audit file from any filesystem.

    Args:
        audit_file_path: Relative path to the audit file from catalog root
        catalog_root: Absolute path to the catalog root directory

    Returns:
        Dictionary containing audit data
    """
    from deltacat.utils.filesystem import resolve_path_and_filesystem
    import json
    import posixpath

    # Resolve absolute path from relative audit path
    absolute_path = posixpath.join(catalog_root, audit_file_path)

    path, filesystem = resolve_path_and_filesystem(absolute_path)
    with filesystem.open_input_stream(path) as stream:
        content = stream.read().decode("utf-8")
        return json.loads(content)



# Allow self-referencing Type hints in Python 3.7.
from __future__ import annotations

from typing import Optional, Any

from deltacat.storage import Stream, Table, TableVersion
from deltacat.storage.model.scan.push_down import Pushdown
from deltacat.storage.model.scan.scan_plan import ScanPlan
from deltacat.storage.util.scan_planner import ScanPlanner


class TableDefinition(dict):
    @staticmethod
    def of(
        table: Table,
        table_version: TableVersion,
        stream: Stream,
        native_object: Optional[Any] = None,
        scan_planner: Optional[ScanPlanner] = None,
    ) -> TableDefinition:
        return TableDefinition(
            {
                "table": table,
                "tableVersion": table_version,
                "stream": stream,
                "nativeObject": native_object,
                "scan_planner": scan_planner,
            }
        )

    @property
    def table(self) -> Table:
        return self["table"]

    @property
    def table_version(self) -> TableVersion:
        return self["tableVersion"]

    @property
    def stream(self) -> Stream:
        return self["stream"]

    @property
    def native_object(self) -> Optional[Any]:
        return self.get("nativeObject")

    @property
    def scan_planner(self) -> Optional[ScanPlanner]:
        return self.get("scan_planner")

    def create_scan_plan(self, pushdown: Optional[Pushdown] = None) -> ScanPlan:
        if not self.scan_planner:
            raise RuntimeError(
                f"ScanPlanner is not initialized for table '{self.table.table_name}' "
                f"of namespace '{self.table.namespace}'"
            )
        return self.scan_planner.create_scan_plan(
            table_name=self.table.table_name,
            namespace=self.table.namespace,
            pushdown=pushdown,
        )



from unittest import TestCase
import numpy as np
import tempfile
import fsspec
import gzip
import polars as pl
from deltacat.types.media import ContentType, ContentEncoding
from deltacat.utils.numpy import (
    file_to_ndarray,
    slice_ndarray,
    ndarray_size,
    ndarray_to_file,
)
from deltacat.utils.pandas import ReadKwargsProviderPandasCsvPureUtf8


class TestNumpyReaders(TestCase):
    def setUp(self):
        # Create test data files for reading
        self.fs = fsspec.filesystem("file")
        self.base_path = tempfile.mkdtemp()
        self.fs.makedirs(self.base_path, exist_ok=True)

        # Create test data as 2D array (3 rows, 3 columns)
        self.expected_data = np.array(
            [["a,b\tc|d", "1", "1.1"], ["e,f\tg|h", "2", "2.2"], ["test", "3", "3.3"]]
        )

        # Write test files in different formats
        self._create_test_files()

    def tearDown(self):
        self.fs.rm(self.base_path, recursive=True)

    def _create_test_files(self):
        # Create CSV file (GZIP compressed)
        csv_path = f"{self.base_path}/test.csv"
        with self.fs.open(csv_path, "wb") as f:
            with gzip.GzipFile(fileobj=f, mode="wb") as gz:
                content = '"a,b\tc|d",1,1.1\n"e,f\tg|h",2,2.2\ntest,3,3.3\n'
                gz.write(content.encode("utf-8"))

        # Create TSV file (GZIP compressed)
        tsv_path = f"{self.base_path}/test.tsv"
        with self.fs.open(tsv_path, "wb") as f:
            with gzip.GzipFile(fileobj=f, mode="wb") as gz:
                content = '"a,b\tc|d"\t1\t1.1\n"e,f\tg|h"\t2\t2.2\ntest\t3\t3.3\n'
                gz.write(content.encode("utf-8"))

        # Create PSV file (GZIP compressed)
        psv_path = f"{self.base_path}/test.psv"
        with self.fs.open(psv_path, "wb") as f:
            with gzip.GzipFile(fileobj=f, mode="wb") as gz:
                content = '"a,b\tc|d"|1|1.1\n"e,f\tg|h"|2|2.2\ntest|3|3.3\n'
                gz.write(content.encode("utf-8"))

        # Create unescaped TSV file (GZIP compressed)
        unescaped_tsv_path = f"{self.base_path}/test_unescaped.tsv"
        with self.fs.open(unescaped_tsv_path, "wb") as f:
            with gzip.GzipFile(fileobj=f, mode="wb") as gz:
                content = "abc\t1\t1.1\ndef\t2\t2.2\nghi\t3\t3.3\n"
                gz.write(content.encode("utf-8"))

        # Create Parquet file
        parquet_path = f"{self.base_path}/test.parquet"
        import pandas as pd

        df = pd.DataFrame(
            {
                "col1": ["a,b\tc|d", "e,f\tg|h", "test"],
                "col2": [1, 2, 3],
                "col3": [1.1, 2.2, 3.3],
            }
        )
        df.to_parquet(parquet_path, index=False)

        # Create Feather file
        feather_path = f"{self.base_path}/test.feather"
        df.to_feather(feather_path)

        # Create JSON file (GZIP compressed, NDJSON format)
        json_path = f"{self.base_path}/test.json"
        with self.fs.open(json_path, "wb") as f:
            with gzip.GzipFile(fileobj=f, mode="wb") as gz:
                json_str = df.to_json(orient="records", lines=True)
                gz.write(json_str.encode("utf-8"))

        # Create Avro file using polars
        avro_path = f"{self.base_path}/test.avro"
        pl_df = pl.from_pandas(df)
        pl_df.write_avro(avro_path)

        # Create ORC file
        orc_path = f"{self.base_path}/test.orc"
        df.to_orc(orc_path, index=False)

    def test_file_to_ndarray_csv(self):
        # Test reading CSV with file_to_ndarray
        csv_path = f"{self.base_path}/test.csv"

        result = file_to_ndarray(
            csv_path,
            ContentType.CSV.value,
            ContentEncoding.GZIP.value,
            filesystem=self.fs,
            column_names=["col1", "col2", "col3"],
        )

        assert result.shape == (3, 3)
        assert result[0, 0] == "a,b\tc|d"
        assert result[1, 0] == "e,f\tg|h"
        assert result[2, 0] == "test"

    def test_file_to_ndarray_tsv(self):
        # Test reading TSV with file_to_ndarray
        tsv_path = f"{self.base_path}/test.tsv"

        result = file_to_ndarray(
            tsv_path,
            ContentType.TSV.value,
            ContentEncoding.GZIP.value,
            filesystem=self.fs,
            column_names=["col1", "col2", "col3"],
        )

        assert result.shape == (3, 3)
        assert result[0, 0] == "a,b\tc|d"
        assert result[1, 0] == "e,f\tg|h"
        assert result[2, 0] == "test"

    def test_file_to_ndarray_psv(self):
        # Test reading PSV with file_to_ndarray
        psv_path = f"{self.base_path}/test.psv"

        result = file_to_ndarray(
            psv_path,
            ContentType.PSV.value,
            ContentEncoding.GZIP.value,
            filesystem=self.fs,
            column_names=["col1", "col2", "col3"],
        )

        assert result.shape == (3, 3)
        assert result[0, 0] == "a,b\tc|d"
        assert result[1, 0] == "e,f\tg|h"
        assert result[2, 0] == "test"

    def test_file_to_ndarray_unescaped_tsv(self):
        # Test reading unescaped TSV with file_to_ndarray
        unescaped_tsv_path = f"{self.base_path}/test_unescaped.tsv"

        result = file_to_ndarray(
            unescaped_tsv_path,
            ContentType.UNESCAPED_TSV.value,
            ContentEncoding.GZIP.value,
            filesystem=self.fs,
            column_names=["col1", "col2", "col3"],
        )

        assert result.shape == (3, 3)
        assert result[0, 0] == "abc"
        assert result[1, 0] == "def"
        assert result[2, 0] == "ghi"

    def test_file_to_ndarray_parquet(self):
        # Test reading Parquet with file_to_ndarray
        parquet_path = f"{self.base_path}/test.parquet"

        result = file_to_ndarray(
            parquet_path, ContentType.PARQUET.value, filesystem=self.fs
        )

        assert result.shape == (3, 3)
        assert result[0, 0] == "a,b\tc|d"
        assert result[1, 0] == "e,f\tg|h"
        assert result[2, 0] == "test"

    def test_file_to_ndarray_feather(self):
        # Test reading Feather with file_to_ndarray
        feather_path = f"{self.base_path}/test.feather"

        result = file_to_ndarray(
            feather_path, ContentType.FEATHER.value, filesystem=self.fs
        )

        assert result.shape == (3, 3)
        assert result[0, 0] == "a,b\tc|d"
        assert result[1, 0] == "e,f\tg|h"
        assert result[2, 0] == "test"

    def test_file_to_ndarray_json(self):
        # Test reading JSON with file_to_ndarray
        json_path = f"{self.base_path}/test.json"

        result = file_to_ndarray(
            json_path,
            ContentType.JSON.value,
            ContentEncoding.GZIP.value,
            filesystem=self.fs,
        )

        assert result.shape == (3, 3)
        # Note: JSON column order might differ, so check by value presence
        assert "a,b\tc|d" in result.flatten()
        assert "e,f\tg|h" in result.flatten()
        assert "test" in result.flatten()

    def test_file_to_ndarray_avro(self):
        # Test reading Avro with file_to_ndarray
        avro_path = f"{self.base_path}/test.avro"

        result = file_to_ndarray(avro_path, ContentType.AVRO.value, filesystem=self.fs)

        assert result.shape == (3, 3)
        assert result[0, 0] == "a,b\tc|d"
        assert result[1, 0] == "e,f\tg|h"
        assert result[2, 0] == "test"

    def test_file_to_ndarray_orc(self):
        # Test reading ORC with file_to_ndarray
        orc_path = f"{self.base_path}/test.orc"

        result = file_to_ndarray(orc_path, ContentType.ORC.value, filesystem=self.fs)

        assert result.shape == (3, 3)
        assert result[0, 0] == "a,b\tc|d"
        assert result[1, 0] == "e,f\tg|h"
        assert result[2, 0] == "test"

    def test_file_to_ndarray_with_column_selection(self):
        # Test reading with column selection
        csv_path = f"{self.base_path}/test.csv"

        result = file_to_ndarray(
            csv_path,
            ContentType.CSV.value,
            ContentEncoding.GZIP.value,
            filesystem=self.fs,
            column_names=["col1", "col2", "col3"],
            include_columns=["col1", "col2"],
        )

        assert result.shape == (3, 2)  # Should only have 2 columns
        assert result[0, 0] == "a,b\tc|d"
        assert result[1, 0] == "e,f\tg|h"
        assert result[2, 0] == "test"

    def test_file_to_ndarray_with_kwargs_provider(self):
        # Test reading with kwargs provider (forces string types)
        csv_path = f"{self.base_path}/test.csv"
        provider = ReadKwargsProviderPandasCsvPureUtf8(
            include_columns=["col1", "col2", "col3"]
        )

        result = file_to_ndarray(
            csv_path,
            ContentType.CSV.value,
            ContentEncoding.GZIP.value,
            filesystem=self.fs,
            column_names=["col1", "col2", "col3"],
            pd_read_func_kwargs_provider=provider,
        )

        assert result.shape == (3, 3)
        assert result[0, 0] == "a,b\tc|d"
        # With string types provider, numbers should also be strings
        assert result[0, 1] == "1"
        assert result[0, 2] == "1.1"

    def test_file_to_ndarray_filesystem_inference(self):
        # Test filesystem inference when no filesystem is provided
        json_path = f"{self.base_path}/test.json"

        result = file_to_ndarray(
            json_path,
            ContentType.JSON.value,
            ContentEncoding.GZIP.value
            # No filesystem provided - should be inferred
        )

        assert result.shape == (3, 3)
        # JSON might have different column ordering
        assert "a,b\tc|d" in result.flatten()
        assert "e,f\tg|h" in result.flatten()
        assert "test" in result.flatten()

    def test_file_to_ndarray_bzip2_compression(self):
        # Test BZIP2 compression handling
        import bz2

        # Create a BZIP2 compressed CSV file
        csv_content = '"a,b\tc|d",1,1.1\n"e,f\tg|h",2,2.2\ntest,3,3.3\n'
        compressed_content = bz2.compress(csv_content.encode("utf-8"))

        bz2_path = f"{self.base_path}/test.csv.bz2"
        with self.fs.open(bz2_path, "wb") as f:
            f.write(compressed_content)

        result = file_to_ndarray(
            bz2_path,
            ContentType.CSV.value,
            ContentEncoding.BZIP2.value,
            filesystem=self.fs,
            column_names=["col1", "col2", "col3"],
        )

        assert result.shape == (3, 3)
        assert result[0, 0] == "a,b\tc|d"
        assert result[1, 0] == "e,f\tg|h"
        assert result[2, 0] == "test"

    def test_slice_ndarray(self):
        # Test slicing functionality
        arr = np.arange(10).reshape(10, 1)

        # Test without max_len (should return original array)
        result = slice_ndarray(arr, None)
        assert len(result) == 1
        np.testing.assert_array_equal(result[0], arr)

        # Test with max_len
        result = slice_ndarray(arr, 3)
        assert len(result) == 4  # 10 rows / 3 = 3 full slices + 1 remainder
        assert result[0].shape == (3, 1)
        assert result[1].shape == (3, 1)
        assert result[2].shape == (3, 1)
        assert result[3].shape == (1, 1)  # remainder

        # Verify data integrity
        np.testing.assert_array_equal(result[0], arr[:3])
        np.testing.assert_array_equal(result[1], arr[3:6])
        np.testing.assert_array_equal(result[2], arr[6:9])
        np.testing.assert_array_equal(result[3], arr[9:])

    def test_ndarray_size(self):
        # Test size calculation
        arr = np.array([[1, 2, 3], [4, 5, 6]], dtype=np.float64)
        size = ndarray_size(arr)
        expected_size = arr.nbytes
        assert size == expected_size

    def test_ndarray_to_file(self):
        # Test writing ndarray to file
        arr = np.array([1, 2, 3, 4, 5])
        path = f"{self.base_path}/test_output.parquet"

        ndarray_to_file(
            arr, path, self.fs, lambda x: path, content_type=ContentType.PARQUET.value
        )

        assert self.fs.exists(path), "file was not written"

        # Verify we can read it back (though this tests the write functionality)
        import pandas as pd

        result_df = pd.read_parquet(path)
        assert len(result_df) == 5
        assert "0" in result_df.columns

    def test_ndarray_to_file_different_content_types(self):
        # Test writing ndarray to different file formats
        arr = np.array([1, 2, 3, 4, 5])

        # Test Parquet
        parquet_path = f"{self.base_path}/test_output.parquet"
        ndarray_to_file(
            arr,
            parquet_path,
            self.fs,
            lambda x: parquet_path,
            content_type=ContentType.PARQUET.value,
        )
        assert self.fs.exists(parquet_path)

        # Test Feather
        feather_path = f"{self.base_path}/test_output.feather"
        ndarray_to_file(
            arr,
            feather_path,
            self.fs,
            lambda x: feather_path,
            content_type=ContentType.FEATHER.value,
        )
        assert self.fs.exists(feather_path)

        # Test CSV (compressed)
        csv_path = f"{self.base_path}/test_output.csv"
        ndarray_to_file(
            arr,
            csv_path,
            self.fs,
            lambda x: csv_path,
            content_type=ContentType.CSV.value,
        )
        assert self.fs.exists(csv_path)

        # Test JSON (compressed)
        json_path = f"{self.base_path}/test_output.json"
        ndarray_to_file(
            arr,
            json_path,
            self.fs,
            lambda x: json_path,
            content_type=ContentType.JSON.value,
        )
        assert self.fs.exists(json_path)

    def test_ndarray_to_file_different_dtypes(self):
        # Test writing arrays with different data types

        # Integer array
        int_arr = np.array([1, 2, 3, 4, 5], dtype=np.int64)
        int_path = f"{self.base_path}/test_int.parquet"
        ndarray_to_file(
            int_arr,
            int_path,
            self.fs,
            lambda x: int_path,
            content_type=ContentType.PARQUET.value,
        )
        assert self.fs.exists(int_path)

        # Float array
        float_arr = np.array([1.1, 2.2, 3.3, 4.4, 5.5], dtype=np.float64)
        float_path = f"{self.base_path}/test_float.parquet"
        ndarray_to_file(
            float_arr,
            float_path,
            self.fs,
            lambda x: float_path,
            content_type=ContentType.PARQUET.value,
        )
        assert self.fs.exists(float_path)

        # String array (object dtype)
        str_arr = np.array(["a", "b", "c", "d", "e"], dtype=object)
        str_path = f"{self.base_path}/test_str.parquet"
        ndarray_to_file(
            str_arr,
            str_path,
            self.fs,
            lambda x: str_path,
            content_type=ContentType.PARQUET.value,
        )
        assert self.fs.exists(str_path)

    def test_ndarray_to_file_2d_array(self):
        # Test writing 2D arrays
        arr_2d = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
        path = f"{self.base_path}/test_2d.parquet"

        ndarray_to_file(
            arr_2d,
            path,
            self.fs,
            lambda x: path,
            content_type=ContentType.PARQUET.value,
        )

        assert self.fs.exists(path)

        # Verify the file structure
        import pandas as pd

        result_df = pd.read_parquet(path)
        assert len(result_df) == 3  # 3 rows (first dimension)
        # 2D array should have columns "0", "1", "2"
        assert list(result_df.columns) == ["0", "1", "2"]
        # Verify the data values are correct (convert back to numpy for comparison)
        result_array = result_df.to_numpy()
        np.testing.assert_array_equal(result_array, arr_2d)

    def test_ndarray_to_file_empty_array(self):
        # Test writing empty arrays
        empty_arr = np.array([])
        path = f"{self.base_path}/test_empty.parquet"

        ndarray_to_file(
            empty_arr,
            path,
            self.fs,
            lambda x: path,
            content_type=ContentType.PARQUET.value,
        )

        assert self.fs.exists(path)

        # Verify the file structure
        import pandas as pd

        result_df = pd.read_parquet(path)
        assert len(result_df) == 0  # Empty DataFrame
        assert "0" in result_df.columns

    def test_ndarray_to_file_large_array(self):
        # Test writing larger arrays
        large_arr = np.arange(1000)
        path = f"{self.base_path}/test_large.parquet"

        ndarray_to_file(
            large_arr,
            path,
            self.fs,
            lambda x: path,
            content_type=ContentType.PARQUET.value,
        )

        assert self.fs.exists(path)

        # Verify the file can be read and has correct size
        import pandas as pd

        result_df = pd.read_parquet(path)
        assert len(result_df) == 1000
        assert "0" in result_df.columns

    def test_ndarray_to_file_with_custom_kwargs(self):
        # Test writing with custom kwargs
        arr = np.array([1, 2, 3, 4, 5])
        path = f"{self.base_path}/test_kwargs.parquet"

        # Add some custom write kwargs (these will be passed to PyArrow)
        ndarray_to_file(
            arr,
            path,
            self.fs,
            lambda x: path,
            content_type=ContentType.PARQUET.value,
            compression="snappy",  # Custom compression
        )

        assert self.fs.exists(path)

        # Verify the file was written (basic check)
        import pandas as pd

        result_df = pd.read_parquet(path)
        assert len(result_df) == 5
        assert "0" in result_df.columns

    def test_ndarray_to_file_readback_verification(self):
        # Test that we can read back the exact data we wrote
        original_arr = np.array([1.1, 2.2, 3.3, 4.4, 5.5])
        path = f"{self.base_path}/test_readback.parquet"

        # Write the array
        ndarray_to_file(
            original_arr,
            path,
            self.fs,
            lambda x: path,
            content_type=ContentType.PARQUET.value,
        )

        # Read it back using pandas and verify content
        import pandas as pd

        result_df = pd.read_parquet(path)
        readback_arr = np.array(result_df["0"].tolist())

        # Check that the data matches
        np.testing.assert_array_almost_equal(original_arr, readback_arr)

    def test_ndarray_to_file_different_filesystems(self):
        # Test with different filesystem implementations
        arr = np.array([1, 2, 3, 4, 5])

        # Test with fsspec filesystem (already used in other tests)
        fsspec_path = f"{self.base_path}/test_fsspec.parquet"
        ndarray_to_file(
            arr,
            fsspec_path,
            self.fs,
            lambda x: fsspec_path,
            content_type=ContentType.PARQUET.value,
        )
        assert self.fs.exists(fsspec_path)

        # Test with None filesystem (should infer local filesystem)
        local_path = f"{self.base_path}/test_local.parquet"
        ndarray_to_file(
            arr,
            local_path,
            None,  # No filesystem specified
            lambda x: local_path,
            content_type=ContentType.PARQUET.value,
        )
        # Check if file exists using the fsspec filesystem
        assert self.fs.exists(local_path)

    def test_ndarray_to_file_boolean_array(self):
        # Test writing boolean arrays
        bool_arr = np.array([True, False, True, False, True])
        path = f"{self.base_path}/test_bool.parquet"

        ndarray_to_file(
            bool_arr,
            path,
            self.fs,
            lambda x: path,
            content_type=ContentType.PARQUET.value,
        )

        assert self.fs.exists(path)

        # Verify the file structure and content
        import pandas as pd

        result_df = pd.read_parquet(path)
        assert len(result_df) == 5
        assert "0" in result_df.columns

        # Check that boolean values are preserved
        readback_arr = np.array(result_df["0"].tolist())
        np.testing.assert_array_equal(bool_arr, readback_arr)

    def test_ndarray_to_file_complex_dtypes(self):
        # Test writing arrays with complex dtypes
        complex_arr = np.array([1 + 2j, 3 + 4j, 5 + 6j])
        path = f"{self.base_path}/test_complex.parquet"

        # Note: Complex numbers might not be directly supported by all formats
        # This test may need to handle conversion or errors gracefully
        try:
            ndarray_to_file(
                complex_arr,
                path,
                self.fs,
                lambda x: path,
                content_type=ContentType.PARQUET.value,
            )
            assert self.fs.exists(path)
        except (TypeError, ValueError, NotImplementedError):
            # Complex dtypes might not be supported by PyArrow/Parquet
            # This is acceptable behavior
            pass


class TestNumpyFileSystemSupport(TestCase):
    """
    Comprehensive tests for numpy file operations with different filesystem types.
    Tests fsspec AbstractFileSystem, PyArrow FileSystem, and auto-inferred filesystem.
    """

    def setUp(self):
        import pyarrow.fs as pafs

        # Create test data as numpy array
        # All formats preserve mixed types when converted to numpy, so use object dtype for all
        self.test_data = np.array(
            [["value1", 1, 1.1], ["value2", 2, 2.2], ["value3", 3, 3.3]], dtype=object
        )

        # Set up temporary directory
        self.temp_dir = tempfile.mkdtemp()

        # Set up different filesystem types
        self.fsspec_fs = fsspec.filesystem("file")
        self.pyarrow_fs = pafs.LocalFileSystem()

        # Create test files for each content type
        self._create_test_files()

    def tearDown(self):
        import shutil

        shutil.rmtree(self.temp_dir)

    def _create_test_files(self):
        """Create test files in different formats with different compression types."""
        import gzip
        import bz2
        import pandas as pd

        # Create pandas DataFrame for file creation
        df = pd.DataFrame(
            {
                "col1": ["value1", "value2", "value3"],
                "col2": [1, 2, 3],
                "col3": [1.1, 2.2, 3.3],
            }
        )

        # CSV files without headers to match test data structure
        csv_data = "value1,1,1.1\nvalue2,2,2.2\nvalue3,3,3.3\n"

        # Create uncompressed CSV
        with open(f"{self.temp_dir}/test.csv", "w") as f:
            f.write(csv_data)

        # Create GZIP compressed CSV
        with gzip.open(f"{self.temp_dir}/test_gzip.csv.gz", "wt") as f:
            f.write(csv_data)

        # Create BZIP2 compressed CSV
        with bz2.open(f"{self.temp_dir}/test_bzip2.csv.bz2", "wt") as f:
            f.write(csv_data)

        # Parquet file
        df.to_parquet(f"{self.temp_dir}/test.parquet", index=False)

        # Feather file
        df.to_feather(f"{self.temp_dir}/test.feather")

        # JSON file (NDJSON format)
        json_str = df.to_json(orient="records", lines=True)
        with open(f"{self.temp_dir}/test.json", "w") as f:
            f.write(json_str)

        # AVRO file (using polars since pandas delegates to polars for AVRO)
        import polars as pl

        pl_df = pl.from_pandas(df)
        pl_df.write_avro(f"{self.temp_dir}/test.avro")

        # ORC file
        df.to_orc(f"{self.temp_dir}/test.orc")

    def _assert_arrays_equal(self, result, expected):
        """Helper to assert numpy arrays are equal."""
        assert (
            result.shape == expected.shape
        ), f"Shape mismatch: {result.shape} vs {expected.shape}"
        np.testing.assert_array_equal(result, expected)

    def test_csv_with_fsspec_filesystem(self):
        """Test CSV reading with fsspec AbstractFileSystem."""
        # Test uncompressed CSV
        result = file_to_ndarray(
            f"{self.temp_dir}/test.csv",
            ContentType.CSV.value,
            ContentEncoding.IDENTITY.value,
            filesystem=self.fsspec_fs,
            column_names=["col1", "col2", "col3"],
        )
        self._assert_arrays_equal(result, self.test_data)

        # Test GZIP compressed CSV
        result = file_to_ndarray(
            f"{self.temp_dir}/test_gzip.csv.gz",
            ContentType.CSV.value,
            ContentEncoding.GZIP.value,
            filesystem=self.fsspec_fs,
            column_names=["col1", "col2", "col3"],
        )
        self._assert_arrays_equal(result, self.test_data)

        # Test BZIP2 compressed CSV
        result = file_to_ndarray(
            f"{self.temp_dir}/test_bzip2.csv.bz2",
            ContentType.CSV.value,
            ContentEncoding.BZIP2.value,
            filesystem=self.fsspec_fs,
            column_names=["col1", "col2", "col3"],
        )
        self._assert_arrays_equal(result, self.test_data)

    def test_csv_with_pyarrow_filesystem(self):
        """Test CSV reading with PyArrow FileSystem."""
        # Test uncompressed CSV
        result = file_to_ndarray(
            f"{self.temp_dir}/test.csv",
            ContentType.CSV.value,
            ContentEncoding.IDENTITY.value,
            filesystem=self.pyarrow_fs,
            column_names=["col1", "col2", "col3"],
        )
        self._assert_arrays_equal(result, self.test_data)

        # Test GZIP compressed CSV
        result = file_to_ndarray(
            f"{self.temp_dir}/test_gzip.csv.gz",
            ContentType.CSV.value,
            ContentEncoding.GZIP.value,
            filesystem=self.pyarrow_fs,
            column_names=["col1", "col2", "col3"],
        )
        self._assert_arrays_equal(result, self.test_data)

    def test_csv_with_auto_inferred_filesystem(self):
        """Test CSV reading with automatically inferred filesystem."""
        # Test uncompressed CSV (filesystem=None, should auto-infer)
        result = file_to_ndarray(
            f"{self.temp_dir}/test.csv",
            ContentType.CSV.value,
            ContentEncoding.IDENTITY.value,
            filesystem=None,
            column_names=["col1", "col2", "col3"],
        )
        self._assert_arrays_equal(result, self.test_data)

    def test_parquet_with_different_filesystems(self):
        """Test Parquet reading with different filesystem types."""
        # Test with fsspec
        result = file_to_ndarray(
            f"{self.temp_dir}/test.parquet",
            ContentType.PARQUET.value,
            ContentEncoding.IDENTITY.value,
            filesystem=self.fsspec_fs,
        )
        self._assert_arrays_equal(result, self.test_data)

        # Test with PyArrow
        result = file_to_ndarray(
            f"{self.temp_dir}/test.parquet",
            ContentType.PARQUET.value,
            ContentEncoding.IDENTITY.value,
            filesystem=self.pyarrow_fs,
        )
        self._assert_arrays_equal(result, self.test_data)

        # Test with auto-inferred
        result = file_to_ndarray(
            f"{self.temp_dir}/test.parquet",
            ContentType.PARQUET.value,
            ContentEncoding.IDENTITY.value,
            filesystem=None,
        )
        self._assert_arrays_equal(result, self.test_data)

    def test_feather_with_different_filesystems(self):
        """Test Feather reading with different filesystem types."""
        # Test with fsspec
        result = file_to_ndarray(
            f"{self.temp_dir}/test.feather",
            ContentType.FEATHER.value,
            ContentEncoding.IDENTITY.value,
            filesystem=self.fsspec_fs,
        )
        self._assert_arrays_equal(result, self.test_data)

        # Test with PyArrow
        result = file_to_ndarray(
            f"{self.temp_dir}/test.feather",
            ContentType.FEATHER.value,
            ContentEncoding.IDENTITY.value,
            filesystem=self.pyarrow_fs,
        )
        self._assert_arrays_equal(result, self.test_data)

        # Test with auto-inferred
        result = file_to_ndarray(
            f"{self.temp_dir}/test.feather",
            ContentType.FEATHER.value,
            ContentEncoding.IDENTITY.value,
            filesystem=None,
        )
        self._assert_arrays_equal(result, self.test_data)

    def test_json_with_different_filesystems(self):
        """Test JSON reading with different filesystem types."""
        # Test with fsspec
        result = file_to_ndarray(
            f"{self.temp_dir}/test.json",
            ContentType.JSON.value,
            ContentEncoding.IDENTITY.value,
            filesystem=self.fsspec_fs,
        )
        self._assert_arrays_equal(result, self.test_data)

        # Test with PyArrow
        result = file_to_ndarray(
            f"{self.temp_dir}/test.json",
            ContentType.JSON.value,
            ContentEncoding.IDENTITY.value,
            filesystem=self.pyarrow_fs,
        )
        self._assert_arrays_equal(result, self.test_data)

        # Test with auto-inferred
        result = file_to_ndarray(
            f"{self.temp_dir}/test.json",
            ContentType.JSON.value,
            ContentEncoding.IDENTITY.value,
            filesystem=None,
        )
        self._assert_arrays_equal(result, self.test_data)

    def test_avro_with_different_filesystems(self):
        """Test AVRO reading with different filesystem types."""
        # Test with fsspec
        result = file_to_ndarray(
            f"{self.temp_dir}/test.avro",
            ContentType.AVRO.value,
            ContentEncoding.IDENTITY.value,
            filesystem=self.fsspec_fs,
        )
        self._assert_arrays_equal(result, self.test_data)

        # Test with PyArrow
        result = file_to_ndarray(
            f"{self.temp_dir}/test.avro",
            ContentType.AVRO.value,
            ContentEncoding.IDENTITY.value,
            filesystem=self.pyarrow_fs,
        )
        self._assert_arrays_equal(result, self.test_data)

        # Test with auto-inferred
        result = file_to_ndarray(
            f"{self.temp_dir}/test.avro",
            ContentType.AVRO.value,
            ContentEncoding.IDENTITY.value,
            filesystem=None,
        )
        self._assert_arrays_equal(result, self.test_data)

    def test_orc_with_different_filesystems(self):
        """Test ORC reading with different filesystem types."""
        # Test with fsspec
        result = file_to_ndarray(
            f"{self.temp_dir}/test.orc",
            ContentType.ORC.value,
            ContentEncoding.IDENTITY.value,
            filesystem=self.fsspec_fs,
        )
        self._assert_arrays_equal(result, self.test_data)

        # Test with PyArrow
        result = file_to_ndarray(
            f"{self.temp_dir}/test.orc",
            ContentType.ORC.value,
            ContentEncoding.IDENTITY.value,
            filesystem=self.pyarrow_fs,
        )
        self._assert_arrays_equal(result, self.test_data)

        # Test with auto-inferred
        result = file_to_ndarray(
            f"{self.temp_dir}/test.orc",
            ContentType.ORC.value,
            ContentEncoding.IDENTITY.value,
            filesystem=None,
        )
        self._assert_arrays_equal(result, self.test_data)

    def test_file_to_ndarray_with_different_filesystems(self):
        """Test file_to_ndarray with different filesystem types for all content types."""
        test_cases = [
            (
                f"{self.temp_dir}/test.csv",
                ContentType.CSV.value,
                ContentEncoding.IDENTITY.value,
                {"column_names": ["col1", "col2", "col3"]},
                self.test_data,
            ),
            (
                f"{self.temp_dir}/test_gzip.csv.gz",
                ContentType.CSV.value,
                ContentEncoding.GZIP.value,
                {"column_names": ["col1", "col2", "col3"]},
                self.test_data,
            ),
            (
                f"{self.temp_dir}/test.parquet",
                ContentType.PARQUET.value,
                ContentEncoding.IDENTITY.value,
                {},
                self.test_data,
            ),
            (
                f"{self.temp_dir}/test.feather",
                ContentType.FEATHER.value,
                ContentEncoding.IDENTITY.value,
                {},
                self.test_data,
            ),
            (
                f"{self.temp_dir}/test.json",
                ContentType.JSON.value,
                ContentEncoding.IDENTITY.value,
                {},
                self.test_data,
            ),
            (
                f"{self.temp_dir}/test.avro",
                ContentType.AVRO.value,
                ContentEncoding.IDENTITY.value,
                {},
                self.test_data,
            ),
            (
                f"{self.temp_dir}/test.orc",
                ContentType.ORC.value,
                ContentEncoding.IDENTITY.value,
                {},
                self.test_data,
            ),
        ]

        filesystems = [
            ("fsspec", self.fsspec_fs),
            ("pyarrow", self.pyarrow_fs),
            ("auto-inferred", None),
        ]

        for (
            path,
            content_type,
            content_encoding,
            extra_kwargs,
            expected_data,
        ) in test_cases:
            for fs_name, filesystem in filesystems:
                with self.subTest(
                    content_type=content_type,
                    filesystem=fs_name,
                    encoding=content_encoding,
                ):
                    result = file_to_ndarray(
                        path=path,
                        content_type=content_type,
                        content_encoding=content_encoding,
                        filesystem=filesystem,
                        **extra_kwargs,
                    )
                    self._assert_arrays_equal(result, expected_data)

    def test_compression_encoding_with_different_filesystems(self):
        """Test that compression encoding works correctly with different filesystem types."""
        test_cases = [
            (f"{self.temp_dir}/test.csv", ContentEncoding.IDENTITY.value),
            (f"{self.temp_dir}/test_gzip.csv.gz", ContentEncoding.GZIP.value),
            (f"{self.temp_dir}/test_bzip2.csv.bz2", ContentEncoding.BZIP2.value),
        ]

        filesystems = [
            ("fsspec", self.fsspec_fs),
            ("pyarrow", self.pyarrow_fs),
            ("auto-inferred", None),
        ]

        for path, content_encoding in test_cases:
            for fs_name, filesystem in filesystems:
                with self.subTest(encoding=content_encoding, filesystem=fs_name):
                    result = file_to_ndarray(
                        path=path,
                        content_type=ContentType.CSV.value,
                        content_encoding=content_encoding,
                        filesystem=filesystem,
                        column_names=["col1", "col2", "col3"],
                    )
                    self._assert_arrays_equal(result, self.test_data)

    def test_filesystem_open_kwargs(self):
        """Test that filesystem open kwargs are properly passed through."""
        # Test with custom fs_open_kwargs
        result = file_to_ndarray(
            f"{self.temp_dir}/test.csv",
            ContentType.CSV.value,
            ContentEncoding.IDENTITY.value,
            filesystem=self.fsspec_fs,
            fs_open_kwargs={
                "encoding": "utf-8"
            },  # This should be passed to filesystem.open()
            column_names=["col1", "col2", "col3"],
        )
        self._assert_arrays_equal(result, self.test_data)

    def test_delimited_formats_with_different_filesystems(self):
        """Test delimited formats (TSV, PSV, etc.) with different filesystem types."""
        # Create TSV test file without headers to match test data structure
        tsv_data = "value1\t1\t1.1\nvalue2\t2\t2.2\nvalue3\t3\t3.3\n"
        with open(f"{self.temp_dir}/test.tsv", "w") as f:
            f.write(tsv_data)

        # Create PSV test file without headers to match test data structure
        psv_data = "value1|1|1.1\nvalue2|2|2.2\nvalue3|3|3.3\n"
        with open(f"{self.temp_dir}/test.psv", "w") as f:
            f.write(psv_data)

        delimited_test_cases = [
            (
                f"{self.temp_dir}/test.tsv",
                ContentType.TSV.value,
                {"sep": "\t", "column_names": ["col1", "col2", "col3"]},
            ),
            (
                f"{self.temp_dir}/test.psv",
                ContentType.PSV.value,
                {"sep": "|", "column_names": ["col1", "col2", "col3"]},
            ),
        ]

        filesystems = [
            ("fsspec", self.fsspec_fs),
            ("pyarrow", self.pyarrow_fs),
            ("auto-inferred", None),
        ]

        for path, content_type, extra_kwargs in delimited_test_cases:
            for fs_name, filesystem in filesystems:
                with self.subTest(content_type=content_type, filesystem=fs_name):
                    result = file_to_ndarray(
                        path=path,
                        content_type=content_type,
                        content_encoding=ContentEncoding.IDENTITY.value,
                        filesystem=filesystem,
                        **extra_kwargs,
                    )
                    self._assert_arrays_equal(result, self.test_data)

    def test_numpy_array_conversion_consistency(self):
        """Test that numpy array conversion is consistent across filesystem types."""
        # Test that the same data produces the same numpy array regardless of filesystem
        filesystems = [
            ("fsspec", self.fsspec_fs),
            ("pyarrow", self.pyarrow_fs),
            ("auto-inferred", None),
        ]

        # Use Parquet as it preserves data types well
        parquet_path = f"{self.temp_dir}/test.parquet"

        results = []
        for fs_name, filesystem in filesystems:
            result = file_to_ndarray(
                parquet_path,
                ContentType.PARQUET.value,
                ContentEncoding.IDENTITY.value,
                filesystem=filesystem,
            )
            results.append((fs_name, result))

        # All results should be identical
        reference_result = results[0][1]
        for fs_name, result in results[1:]:
            with self.subTest(filesystem=fs_name):
                self._assert_arrays_equal(result, reference_result)

    def test_dtype_preservation_across_filesystems(self):
        """Test that data types are preserved across different filesystem types."""
        import pandas as pd

        # Create a DataFrame with mixed data types
        df = pd.DataFrame(
            {
                "int_col": [1, 2, 3],
                "float_col": [1.1, 2.2, 3.3],
                "str_col": ["a", "b", "c"],
            }
        )

        # Save as Parquet (preserves types best)
        parquet_path = f"{self.temp_dir}/test_dtypes.parquet"
        df.to_parquet(parquet_path, index=False)

        filesystems = [
            ("fsspec", self.fsspec_fs),
            ("pyarrow", self.pyarrow_fs),
            ("auto-inferred", None),
        ]

        # Test that data types are consistent across filesystems
        dtypes = []
        for fs_name, filesystem in filesystems:
            result = file_to_ndarray(
                parquet_path,
                ContentType.PARQUET.value,
                ContentEncoding.IDENTITY.value,
                filesystem=filesystem,
            )
            dtypes.append((fs_name, result.dtype))

        # All dtypes should be the same (object type for mixed data)
        reference_dtype = dtypes[0][1]
        for fs_name, dtype in dtypes[1:]:
            with self.subTest(filesystem=fs_name):
                assert (
                    dtype == reference_dtype
                ), f"Dtype mismatch for {fs_name}: {dtype} vs {reference_dtype}"

    def test_error_handling_across_filesystems(self):
        """Test that error handling is consistent across filesystem types."""
        filesystems = [
            ("fsspec", self.fsspec_fs),
            ("pyarrow", self.pyarrow_fs),
            ("auto-inferred", None),
        ]

        # Test with non-existent file
        for fs_name, filesystem in filesystems:
            with self.subTest(filesystem=fs_name):
                with self.assertRaises(
                    Exception
                ):  # Should raise some kind of file not found error
                    file_to_ndarray(
                        f"{self.temp_dir}/nonexistent.csv",
                        ContentType.CSV.value,
                        ContentEncoding.IDENTITY.value,
                        filesystem=filesystem,
                        column_names=["col1", "col2", "col3"],
                    )



import logging
import ray
import functools
from typing import List
from deltacat.compute.compactor_v2.constants import (
    TASK_MAX_PARALLELISM,
    MAX_PARQUET_METADATA_SIZE,
)
from deltacat.utils.common import ReadKwargsProvider
from deltacat.utils.ray_utils.concurrency import invoke_parallel
from deltacat import logs
from deltacat.storage import (
    Delta,
    ManifestEntry,
    metastore,
)
from typing import Dict, Optional, Any
from deltacat.types.media import DatasetType
from deltacat.types.media import ContentType
from deltacat.types.partial_download import PartialParquetParameters
from deltacat.exceptions import RetryableError

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def append_content_type_params_options_provider(
    index: int, item: Any, max_parquet_meta_size_bytes: int, **kwargs
) -> Dict:
    task_opts = {
        "num_cpus": 0.01,
        "memory": max_parquet_meta_size_bytes,
        "scheduling_strategy": "DEFAULT",
    }

    task_opts["max_retries"] = 3
    task_opts["retry_exceptions"] = [RetryableError]

    return task_opts


def _contains_partial_parquet_parameters(entry: ManifestEntry) -> bool:
    return (
        entry.meta
        and entry.meta.content_type_parameters
        and any(
            isinstance(type_params, PartialParquetParameters)
            for type_params in entry.meta.content_type_parameters
        )
    )


APPEND_CONTENT_TYPE_PARAMS_CACHE = "append_content_type_params_cache"
# At this point, it's better to fetch all parquet than to cache and
# call actor which is not expected to support high throughput.
MINIMUM_ENTRIES_TO_CACHE = 10


@ray.remote
class AppendContentTypeParamsCache:
    """
    This actor caches the delta that contains content type meta.
    """

    def __init__(self):
        self.cache = {}

    def get(self, key):
        return self.cache.get(key)

    def put(self, key, value):
        self.cache[key] = value


@ray.remote
def _download_parquet_metadata_for_manifest_entry(
    delta: Delta,
    entry_index: int,
    all_column_names: List[str],
    deltacat_storage: metastore,
    deltacat_storage_kwargs: Optional[Dict[Any, Any]] = {},
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
) -> Dict[str, Any]:
    logger.info(
        f"Downloading the parquet metadata for Delta with locator {delta.locator} and entry_index: {entry_index}"
    )
    if "file_reader_kwargs_provider" in deltacat_storage_kwargs:
        logger.info(
            "'file_reader_kwargs_provider' is also present in deltacat_storage_kwargs. Removing to prevent multiple values for keyword argument"
        )
        deltacat_storage_kwargs.pop("file_reader_kwargs_provider")

    pq_file = deltacat_storage.download_delta_manifest_entry(
        delta,
        entry_index=entry_index,
        table_type=DatasetType.PYARROW_PARQUET,
        file_reader_kwargs_provider=file_reader_kwargs_provider,
        all_column_names=all_column_names,
        **deltacat_storage_kwargs,
    )

    return {
        "entry_index": entry_index,
        "partial_parquet_params": PartialParquetParameters.of(
            pq_metadata=pq_file.metadata
        ),
    }


def append_content_type_params(
    delta: Delta,
    all_column_names: List[str],
    task_max_parallelism: int = TASK_MAX_PARALLELISM,
    max_parquet_meta_size_bytes: Optional[int] = MAX_PARQUET_METADATA_SIZE,
    deltacat_storage: metastore = metastore,
    deltacat_storage_kwargs: Optional[Dict[str, Any]] = {},
    file_reader_kwargs_provider: Optional[ReadKwargsProvider] = None,
) -> bool:
    """
    This operation appends content type params into the delta entry. Note
    that this operation can be time consuming, hence we cache it in a Ray actor.
    """
    logger.info(
        f"Appending the content type params for Delta with locator {delta.locator}..."
    )

    if not delta.meta:
        logger.warning(f"Delta with locator {delta.locator} doesn't contain meta.")
        return False

    entry_indices_to_download = []
    for entry_index, entry in enumerate(delta.manifest.entries):
        if (
            not _contains_partial_parquet_parameters(entry)
            and entry.meta
            and entry.meta.content_type == ContentType.PARQUET.value
        ):
            entry_indices_to_download.append(entry_index)

    if not entry_indices_to_download:
        logger.info(
            f"No parquet type params to download for delta with locator {delta.locator}."
        )
        return False

    ray_namespace = ray.get_runtime_context().namespace
    logger.info(
        f"Got Ray namespace: {ray_namespace}. "
        "Note that caching only works with non-anonymous namespace."
        "To set a non-anonymous namespace, call ray.init(namespace='X')."
    )
    if len(entry_indices_to_download) >= MINIMUM_ENTRIES_TO_CACHE:
        logger.info(
            f"Checking if cache contains parquet meta in namespace {ray_namespace} for "
            f"delta locator {delta.locator} and digest {delta.locator.hexdigest()}..."
        )
        cache = AppendContentTypeParamsCache.options(
            name=APPEND_CONTENT_TYPE_PARAMS_CACHE,
            namespace=ray_namespace,
            get_if_exists=True,
        ).remote()

        logger.info(f"Got cache actor: {cache}")
        cached_value = ray.get(cache.get.remote(delta.locator.hexdigest()))
        if cached_value is not None:
            logger.info(
                "Using cached parquet meta for delta with locator"
                f" {delta.locator} and digest {delta.locator.hexdigest()}."
            )
            delta.manifest = cached_value.manifest
            return True
        logger.info(
            f"Cache doesn't contain parquet meta for delta with locator {delta.locator}."
        )

    options_provider = functools.partial(
        append_content_type_params_options_provider,
        max_parquet_meta_size_bytes=max_parquet_meta_size_bytes,
    )

    # create a copy of deltacat_storage_kwargs without transaction key
    deltacat_storage_kwargs_copy = {
        k: v for k, v in deltacat_storage_kwargs.items() if k != "transaction"
    }

    def input_provider(index, item) -> Dict:
        return {
            "file_reader_kwargs_provider": file_reader_kwargs_provider,
            "deltacat_storage_kwargs": deltacat_storage_kwargs_copy,
            "deltacat_storage": deltacat_storage,
            "delta": delta,
            "entry_index": item,
            "all_column_names": all_column_names,
        }

    logger.info(
        f"Downloading parquet meta for {len(entry_indices_to_download)} manifest entries..."
    )

    pq_files_promise = invoke_parallel(
        entry_indices_to_download,
        ray_task=_download_parquet_metadata_for_manifest_entry,
        max_parallelism=task_max_parallelism,
        options_provider=options_provider,
        kwargs_provider=input_provider,
    )

    partial_file_meta_list = ray.get(pq_files_promise)

    logger.info(
        f"Downloaded parquet meta for {len(entry_indices_to_download)} manifest entries"
    )

    assert len(partial_file_meta_list) == len(
        entry_indices_to_download
    ), f"Expected {len(entry_indices_to_download)} pq files, got {len(partial_file_meta_list)}"

    for index, entry_index in enumerate(entry_indices_to_download):
        assert (
            entry_index == partial_file_meta_list[index]["entry_index"]
        ), "entry_index must match with the associated parquet meta"
        entry = delta.manifest.entries[entry_index]
        if not entry.meta.content_type_parameters:
            entry.meta.content_type_parameters = []
        entry.meta.content_type_parameters.append(
            partial_file_meta_list[index]["partial_parquet_params"]
        )

    for entry_index, entry in enumerate(delta.manifest.entries):
        assert _contains_partial_parquet_parameters(
            entry
        ), "partial parquet params validation failed."

    if len(entry_indices_to_download) >= MINIMUM_ENTRIES_TO_CACHE:
        cache = AppendContentTypeParamsCache.options(
            name=APPEND_CONTENT_TYPE_PARAMS_CACHE,
            namespace=ray_namespace,
            get_if_exists=True,
        ).remote()
        logger.info(f"Got cache actor when writing: {cache}")
        logger.info(
            f"Caching parquet meta for delta with locator {delta.locator} "
            f"and digest {delta.locator.hexdigest()}..."
        )
        ray.get(cache.put.remote(delta.locator.hexdigest(), delta))
        assert ray.get(cache.get.remote(delta.locator.hexdigest())) is not None

    return True



import logging
import bz2
import gzip
from functools import partial
from typing import Optional, List, Dict, Callable, Union, Iterable, Any

import polars as pl
import pyarrow as pa
import pyarrow.fs as pafs

from fsspec import AbstractFileSystem
from ray.data.datasource import FilenameProvider

from deltacat import logs
from deltacat.utils.filesystem import resolve_path_and_filesystem
from deltacat.utils.common import ContentTypeKwargsProvider, ReadKwargsProvider
from deltacat.utils.performance import timed_invocation

from deltacat.types.media import (
    ContentType,
    ContentEncoding,
    DELIMITED_TEXT_CONTENT_TYPES,
    TABULAR_CONTENT_TYPES,
)
from deltacat.types.partial_download import PartialFileDownloadParams

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))

# Encoding to file initialization function mapping
ENCODING_TO_FILE_INIT: Dict[str, Callable] = {
    ContentEncoding.GZIP.value: partial(gzip.open, mode="rb"),
    ContentEncoding.BZIP2.value: partial(bz2.open, mode="rb"),
    ContentEncoding.IDENTITY.value: lambda file_path: file_path,
}


def write_json(
    table: pl.DataFrame,
    path: str,
    *,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]] = None,
    fs_open_kwargs: Dict[str, any] = {},
    **write_kwargs,
) -> None:
    # Check if the path already indicates compression to avoid double compression
    should_compress = path.endswith(".gz")

    if not filesystem or isinstance(filesystem, pafs.FileSystem):
        path, filesystem = resolve_path_and_filesystem(path, filesystem)
        with filesystem.open_output_stream(path, **fs_open_kwargs) as f:
            if should_compress:
                # Path ends with .gz, PyArrow filesystem automatically compresses
                table.write_ndjson(f, **write_kwargs)
            else:
                # No compression indicated, write uncompressed
                table.write_ndjson(f, **write_kwargs)
    else:
        with filesystem.open(path, "wb", **fs_open_kwargs) as f:
            if should_compress:
                # For fsspec filesystems, we need to apply compression explicitly
                with pa.CompressedOutputStream(f, ContentEncoding.GZIP.value) as out:
                    table.write_ndjson(out, **write_kwargs)
            else:
                # No compression indicated, write uncompressed
                table.write_ndjson(f, **write_kwargs)


def content_type_to_writer_kwargs(content_type: str) -> Dict[str, any]:
    """
    Returns writer kwargs for the given content type when writing with polars.
    """
    if content_type == ContentType.UNESCAPED_TSV.value:
        return {
            "separator": "\t",
            "include_header": False,
            "null_value": "",
            "quote_style": "never",  # Equivalent to QUOTE_NONE in pandas
        }
    if content_type == ContentType.TSV.value:
        return {
            "separator": "\t",
            "include_header": False,
            "quote_style": "necessary",
        }
    if content_type == ContentType.CSV.value:
        return {
            "separator": ",",
            "include_header": False,
            "quote_style": "necessary",
        }
    if content_type == ContentType.PSV.value:
        return {
            "separator": "|",
            "include_header": False,
            "quote_style": "necessary",
        }
    if content_type in {
        ContentType.PARQUET.value,
        ContentType.FEATHER.value,
        ContentType.JSON.value,
        ContentType.AVRO.value,
        ContentType.ORC.value,
    }:
        return {}
    raise ValueError(f"Unsupported content type: {content_type}")


def write_csv(
    table: pl.DataFrame,
    path: str,
    *,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]] = None,
    fs_open_kwargs: Dict[str, any] = {},
    **kwargs,
) -> None:
    """
    Write a polars DataFrame to a CSV file (or other delimited text format).
    """
    # Check if the path already indicates compression to avoid double compression
    should_compress = path.endswith(".gz")

    if not filesystem or isinstance(filesystem, pafs.FileSystem):
        path, filesystem = resolve_path_and_filesystem(path, filesystem)
        with filesystem.open_output_stream(path, **fs_open_kwargs) as f:
            if should_compress:
                # Path ends with .gz, PyArrow filesystem automatically compresses
                table.write_csv(f, **kwargs)
            else:
                # No compression indicated, write uncompressed
                table.write_csv(f, **kwargs)
    else:
        with filesystem.open(path, "wb", **fs_open_kwargs) as f:
            if should_compress:
                # For fsspec filesystems, we need to apply compression explicitly
                with pa.CompressedOutputStream(f, ContentEncoding.GZIP.value) as out:
                    table.write_csv(out, **kwargs)
            else:
                # No compression indicated, write uncompressed
                table.write_csv(f, **kwargs)


def write_avro(
    table: pl.DataFrame,
    path: str,
    *,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]] = None,
    fs_open_kwargs: Dict[str, any] = {},
    **write_kwargs,
) -> None:
    if not filesystem or isinstance(filesystem, pafs.FileSystem):
        path, filesystem = resolve_path_and_filesystem(path, filesystem)
        with filesystem.open_output_stream(path, **fs_open_kwargs) as f:
            table.write_avro(f, **write_kwargs)
    else:
        with filesystem.open(path, "wb", **fs_open_kwargs) as f:
            table.write_avro(f, **write_kwargs)


def write_parquet(
    table: pl.DataFrame,
    path: str,
    *,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]] = None,
    fs_open_kwargs: Dict[str, any] = {},
    **write_kwargs,
) -> None:
    if not filesystem or isinstance(filesystem, pafs.FileSystem):
        path, filesystem = resolve_path_and_filesystem(path, filesystem)
        with filesystem.open_output_stream(path, **fs_open_kwargs) as f:
            table.write_parquet(f, **write_kwargs)
    else:
        with filesystem.open(path, "wb", **fs_open_kwargs) as f:
            table.write_parquet(f, **write_kwargs)


def write_feather(
    table: pl.DataFrame,
    path: str,
    *,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]] = None,
    fs_open_kwargs: Dict[str, any] = {},
    **kwargs,
) -> None:
    """
    Write a polars DataFrame to a Feather file.
    """
    if not filesystem or isinstance(filesystem, pafs.FileSystem):
        path, filesystem = resolve_path_and_filesystem(path, filesystem)
        with filesystem.open_output_stream(path, **fs_open_kwargs) as f:
            table.write_ipc(f, **kwargs)
    else:
        with filesystem.open(path, "wb", **fs_open_kwargs) as f:
            table.write_ipc(f, **kwargs)


def write_orc(
    table: pl.DataFrame,
    path: str,
    *,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]] = None,
    fs_open_kwargs: Dict[str, any] = {},
    **write_kwargs,
) -> None:
    """
    Write a polars DataFrame to an ORC file by delegating to PyArrow implementation.
    """
    from deltacat.utils.pyarrow import write_orc as pyarrow_write_orc

    # Convert polars DataFrame to PyArrow Table
    pa_table = table.to_arrow()

    # Delegate to PyArrow write_orc implementation
    pyarrow_write_orc(
        pa_table,
        path,
        filesystem=filesystem,
        fs_open_kwargs=fs_open_kwargs,
        **write_kwargs,
    )


CONTENT_TYPE_TO_PL_WRITE_FUNC: Dict[str, Callable] = {
    ContentType.UNESCAPED_TSV.value: write_csv,
    ContentType.TSV.value: write_csv,
    ContentType.CSV.value: write_csv,
    ContentType.PSV.value: write_csv,
    ContentType.PARQUET.value: write_parquet,
    ContentType.FEATHER.value: write_feather,
    ContentType.JSON.value: write_json,
    ContentType.AVRO.value: write_avro,
    ContentType.ORC.value: write_orc,
}


def slice_table(table: pl.DataFrame, max_len: Optional[int]) -> List[pl.DataFrame]:
    """
    Iteratively create 0-copy table slices.
    """
    if max_len is None:
        return [table]
    tables = []
    offset = 0
    records_remaining = len(table)
    while records_remaining > 0:
        records_this_entry = min(max_len, records_remaining)
        tables.append(table.slice(offset, records_this_entry))
        records_remaining -= records_this_entry
        offset += records_this_entry
    return tables


def dataframe_size(table: pl.DataFrame) -> int:
    return table.estimated_size()


def dataframe_to_file(
    table: pl.DataFrame,
    base_path: str,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]],
    block_path_provider: Union[Callable, FilenameProvider],
    content_type: str = ContentType.PARQUET.value,
    schema: Optional[pa.Schema] = None,
    **kwargs,
) -> None:
    """
    Writes the given Polars DataFrame to a file.
    """
    writer = CONTENT_TYPE_TO_PL_WRITE_FUNC.get(content_type)
    writer_kwargs = content_type_to_writer_kwargs(content_type)
    writer_kwargs.update(kwargs)
    if not writer:
        raise NotImplementedError(
            f"Polars writer for content type '{content_type}' not "
            f"implemented. Known content types: "
            f"{CONTENT_TYPE_TO_PL_WRITE_FUNC.keys()}"
        )
    path = block_path_provider(base_path)
    logger.debug(f"Writing table: {table} with kwargs: {writer_kwargs} to path: {path}")
    writer(table, path, filesystem=filesystem, **writer_kwargs)


def write_table(
    table: pl.DataFrame,
    path: str,
    *,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]] = None,
    fs_open_kwargs: Dict[str, any] = {},
    content_type: str = ContentType.PARQUET.value,
    **kwargs,
) -> None:
    """
    Write a polars DataFrame to a file in the specified format.
    """
    writer = CONTENT_TYPE_TO_PL_WRITE_FUNC.get(content_type)
    writer_kwargs = content_type_to_writer_kwargs(content_type)
    writer_kwargs.update(kwargs)
    if not writer:
        raise NotImplementedError(
            f"Polars writer for content type '{content_type}' not "
            f"implemented. Known content types: "
            f"{CONTENT_TYPE_TO_PL_WRITE_FUNC.keys()}"
        )
    writer(
        table,
        path,
        filesystem=filesystem,
        fs_open_kwargs=fs_open_kwargs,
        **writer_kwargs,
    )


CONTENT_TYPE_TO_PL_READ_FUNC: Dict[str, Callable] = {
    ContentType.UNESCAPED_TSV.value: pl.read_csv,
    ContentType.TSV.value: pl.read_csv,
    ContentType.CSV.value: pl.read_csv,
    ContentType.PSV.value: pl.read_csv,
    ContentType.PARQUET.value: pl.read_parquet,
    ContentType.FEATHER.value: pl.read_ipc,
    ContentType.JSON.value: pl.read_ndjson,
    ContentType.AVRO.value: pl.read_avro,
}


class ReadKwargsProviderPolarsStringTypes(ContentTypeKwargsProvider):
    """ReadKwargsProvider impl that reads columns of delimited text files
    as UTF-8 strings (i.e. disables type inference). Useful for ensuring
    lossless reads of UTF-8 delimited text datasets and improving read
    performance in cases where type casting is not required."""

    def __init__(self, include_columns: Optional[Iterable[str]] = None):
        self.include_columns = include_columns

    def _get_kwargs(self, content_type: str, kwargs: Dict[str, Any]) -> Dict[str, Any]:
        if content_type in DELIMITED_TEXT_CONTENT_TYPES:
            include_columns = (
                self.include_columns if self.include_columns else kwargs.get("columns")
            )
            if not include_columns:
                # read all columns as strings - disable schema inference
                kwargs["infer_schema"] = False
            else:
                # read only the included columns as strings
                kwargs["schema_overrides"] = {
                    column_name: pl.Utf8 for column_name in include_columns
                }
        return kwargs


def content_type_to_reader_kwargs(content_type: str) -> Dict[str, Any]:
    if content_type == ContentType.UNESCAPED_TSV.value:
        return {
            "separator": "\t",
            "has_header": False,
            "null_values": [""],
            "quote_char": None,
        }
    if content_type == ContentType.TSV.value:
        return {"separator": "\t", "has_header": False}
    if content_type == ContentType.CSV.value:
        return {"separator": ",", "has_header": False}
    if content_type == ContentType.PSV.value:
        return {"separator": "|", "has_header": False}
    if content_type in {
        ContentType.PARQUET.value,
        ContentType.FEATHER.value,
        ContentType.ORC.value,
        ContentType.JSON.value,
        ContentType.AVRO.value,
    }:
        return {}
    raise ValueError(f"Unsupported content type: {content_type}")


def _add_column_kwargs(
    content_type: str,
    column_names: Optional[List[str]],
    include_columns: Optional[List[str]],
    kwargs: Dict[str, Any],
):
    if content_type in DELIMITED_TEXT_CONTENT_TYPES:
        if column_names:
            kwargs["new_columns"] = column_names
        if include_columns:
            kwargs["columns"] = include_columns
    else:
        if content_type in TABULAR_CONTENT_TYPES:
            if include_columns:
                kwargs["columns"] = include_columns
        else:
            if include_columns:
                logger.warning(
                    f"Ignoring request to include columns {include_columns} "
                    f"for non-tabular content type {content_type}"
                )


def concat_dataframes(dataframes: List[pl.DataFrame]) -> Optional[pl.DataFrame]:
    if dataframes is None or not len(dataframes):
        return None
    if len(dataframes) == 1:
        return next(iter(dataframes))
    return pl.concat(dataframes)


def append_column_to_table(
    table: pl.DataFrame,
    column_name: str,
    column_value: Any,
) -> pl.DataFrame:
    return table.with_columns(pl.lit(column_value).alias(column_name))


def select_columns(
    table: pl.DataFrame,
    column_names: List[str],
) -> pl.DataFrame:
    return table.select(column_names)


def file_to_dataframe(
    path: str,
    content_type: str,
    content_encoding: str = ContentEncoding.IDENTITY.value,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]] = None,
    column_names: Optional[List[str]] = None,
    include_columns: Optional[List[str]] = None,
    pl_read_func_kwargs_provider: Optional[ReadKwargsProvider] = None,
    partial_file_download_params: Optional[PartialFileDownloadParams] = None,
    fs_open_kwargs: Dict[str, Any] = {},
    **kwargs,
) -> pl.DataFrame:
    """
    Read a file into a Polars DataFrame using any filesystem.

    Args:
        path: The file path to read
        content_type: The content type of the file (e.g., ContentType.CSV.value)
        content_encoding: The content encoding (default: IDENTITY)
        filesystem: The filesystem to use (if None, will be inferred from path)
        column_names: Optional column names to assign
        include_columns: Optional columns to include in the result
        pl_read_func_kwargs_provider: Optional kwargs provider for customization
        fs_open_kwargs: Optional kwargs for filesystem open operations
        **kwargs: Additional kwargs passed to the reader function

    Returns:
        pl.DataFrame: The loaded DataFrame
    """
    logger.debug(
        f"Reading {path} to Polars. Content type: {content_type}. "
        f"Encoding: {content_encoding}"
    )

    pl_read_func = CONTENT_TYPE_TO_READ_FN.get(content_type)
    if not pl_read_func:
        raise NotImplementedError(
            f"Polars reader for content type '{content_type}' not "
            f"implemented. Known content types: "
            f"{list(CONTENT_TYPE_TO_READ_FN.keys())}"
        )

    reader_kwargs = content_type_to_reader_kwargs(content_type)
    _add_column_kwargs(content_type, column_names, include_columns, reader_kwargs)

    # Merge with provided kwargs
    reader_kwargs.update(kwargs)

    if pl_read_func_kwargs_provider:
        reader_kwargs = pl_read_func_kwargs_provider(content_type, reader_kwargs)

    logger.debug(f"Reading {path} via {pl_read_func} with kwargs: {reader_kwargs}")

    dataframe, latency = timed_invocation(
        pl_read_func,
        path,
        filesystem=filesystem,
        fs_open_kwargs=fs_open_kwargs,
        content_encoding=content_encoding,
        **reader_kwargs,
    )
    logger.debug(f"Time to read {path} into Polars DataFrame: {latency}s")
    return dataframe


def read_csv(
    path: str,
    *,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]] = None,
    fs_open_kwargs: Dict[str, any] = {},
    content_encoding: str = ContentEncoding.IDENTITY.value,
    **read_kwargs,
) -> pl.DataFrame:
    if not filesystem or isinstance(filesystem, pafs.FileSystem):
        path, filesystem = resolve_path_and_filesystem(path)
        if content_encoding == ContentEncoding.IDENTITY.value:
            with filesystem.open_input_stream(path, **fs_open_kwargs) as f:
                return pl.read_csv(f, **read_kwargs)
        else:
            # For compressed files with PyArrow, we need to be careful because PyArrow
            # may auto-decompress some formats. Try to read directly first.
            try:
                with filesystem.open_input_stream(path, **fs_open_kwargs) as f:
                    # Try reading as if it's already decompressed by PyArrow
                    return pl.read_csv(f, **read_kwargs)
            except Exception:
                # If that fails, try manual decompression
                with filesystem.open_input_file(path, **fs_open_kwargs) as f:
                    input_file_init = ENCODING_TO_FILE_INIT.get(
                        content_encoding, lambda x: x
                    )
                    with input_file_init(f) as input_file:
                        content = input_file.read()
                        if isinstance(content, str):
                            content = content.encode("utf-8")
                        return pl.read_csv(content, **read_kwargs)
    else:
        # fsspec AbstractFileSystem
        with filesystem.open(path, "rb", **fs_open_kwargs) as f:
            # Handle compression
            if content_encoding == ContentEncoding.IDENTITY.value:
                return pl.read_csv(f, **read_kwargs)
            else:
                input_file_init = ENCODING_TO_FILE_INIT.get(
                    content_encoding, lambda x: x
                )
                with input_file_init(f) as input_file:
                    # Read decompressed content as bytes and pass to polars
                    content = input_file.read()
                    if isinstance(content, str):
                        content = content.encode("utf-8")
                    return pl.read_csv(content, **read_kwargs)


def read_parquet(
    path: str,
    *,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]] = None,
    fs_open_kwargs: Dict[str, any] = {},
    content_encoding: str = ContentEncoding.IDENTITY.value,
    **read_kwargs,
) -> pl.DataFrame:
    if not filesystem or isinstance(filesystem, pafs.FileSystem):
        path, filesystem = resolve_path_and_filesystem(path)
        with filesystem.open_input_file(path, **fs_open_kwargs) as f:
            # Handle compression
            if content_encoding == ContentEncoding.IDENTITY.value:
                return pl.read_parquet(f, **read_kwargs)
            else:
                input_file_init = ENCODING_TO_FILE_INIT.get(
                    content_encoding, lambda x: x
                )
                with input_file_init(f) as input_file:
                    # Read decompressed content as bytes and pass to polars
                    content = input_file.read()
                    return pl.read_parquet(content, **read_kwargs)
    else:
        # fsspec AbstractFileSystem
        with filesystem.open(path, "rb", **fs_open_kwargs) as f:
            # Handle compression
            if content_encoding == ContentEncoding.IDENTITY.value:
                return pl.read_parquet(f, **read_kwargs)
            else:
                input_file_init = ENCODING_TO_FILE_INIT.get(
                    content_encoding, lambda x: x
                )
                with input_file_init(f) as input_file:
                    # Read decompressed content as bytes and pass to polars
                    content = input_file.read()
                    return pl.read_parquet(content, **read_kwargs)


def read_ipc(
    path: str,
    *,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]] = None,
    fs_open_kwargs: Dict[str, any] = {},
    content_encoding: str = ContentEncoding.IDENTITY.value,
    **read_kwargs,
) -> pl.DataFrame:
    if not filesystem or isinstance(filesystem, pafs.FileSystem):
        path, filesystem = resolve_path_and_filesystem(path)
        with filesystem.open_input_file(path, **fs_open_kwargs) as f:
            # Handle compression
            if content_encoding == ContentEncoding.IDENTITY.value:
                return pl.read_ipc(f, **read_kwargs)
            else:
                input_file_init = ENCODING_TO_FILE_INIT.get(
                    content_encoding, lambda x: x
                )
                with input_file_init(f) as input_file:
                    # Read decompressed content as bytes and pass to polars
                    content = input_file.read()
                    return pl.read_ipc(content, **read_kwargs)
    else:
        # fsspec AbstractFileSystem
        with filesystem.open(path, "rb", **fs_open_kwargs) as f:
            # Handle compression
            if content_encoding == ContentEncoding.IDENTITY.value:
                return pl.read_ipc(f, **read_kwargs)
            else:
                input_file_init = ENCODING_TO_FILE_INIT.get(
                    content_encoding, lambda x: x
                )
                with input_file_init(f) as input_file:
                    # Read decompressed content as bytes and pass to polars
                    content = input_file.read()
                    return pl.read_ipc(content, **read_kwargs)


def read_ndjson(
    path: str,
    *,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]] = None,
    fs_open_kwargs: Dict[str, any] = {},
    content_encoding: str = ContentEncoding.IDENTITY.value,
    **read_kwargs,
) -> pl.DataFrame:
    if not filesystem or isinstance(filesystem, pafs.FileSystem):
        path, filesystem = resolve_path_and_filesystem(path)
        if content_encoding == ContentEncoding.IDENTITY.value:
            with filesystem.open_input_stream(path, **fs_open_kwargs) as f:
                return pl.read_ndjson(f, **read_kwargs)
        else:
            # For compressed files with PyArrow, we need to be careful because PyArrow
            # may auto-decompress some formats. Try to read directly first.
            try:
                with filesystem.open_input_stream(path, **fs_open_kwargs) as f:
                    # Try reading as if it's already decompressed by PyArrow
                    return pl.read_ndjson(f, **read_kwargs)
            except Exception:
                # If that fails, try manual decompression
                with filesystem.open_input_file(path, **fs_open_kwargs) as f:
                    input_file_init = ENCODING_TO_FILE_INIT.get(
                        content_encoding, lambda x: x
                    )
                    with input_file_init(f) as input_file:
                        content = input_file.read()
                        if isinstance(content, str):
                            content = content.encode("utf-8")
                        return pl.read_ndjson(content, **read_kwargs)
    else:
        # fsspec AbstractFileSystem
        with filesystem.open(path, "rb", **fs_open_kwargs) as f:
            # Handle compression
            if content_encoding == ContentEncoding.IDENTITY.value:
                return pl.read_ndjson(f, **read_kwargs)
            else:
                input_file_init = ENCODING_TO_FILE_INIT.get(
                    content_encoding, lambda x: x
                )
                with input_file_init(f) as input_file:
                    # Read decompressed content as bytes and pass to polars
                    content = input_file.read()
                    if isinstance(content, str):
                        content = content.encode("utf-8")
                    return pl.read_ndjson(content, **read_kwargs)


def read_avro(
    path: str,
    *,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]] = None,
    fs_open_kwargs: Dict[str, any] = {},
    content_encoding: str = ContentEncoding.IDENTITY.value,
    **read_kwargs,
) -> pl.DataFrame:
    if not filesystem or isinstance(filesystem, pafs.FileSystem):
        path, filesystem = resolve_path_and_filesystem(path)
        with filesystem.open_input_file(path, **fs_open_kwargs) as f:
            # Handle compression
            if content_encoding == ContentEncoding.IDENTITY.value:
                return pl.read_avro(f, **read_kwargs)
            else:
                input_file_init = ENCODING_TO_FILE_INIT.get(
                    content_encoding, lambda x: x
                )
                with input_file_init(f) as input_file:
                    # Read decompressed content as bytes and pass to polars
                    content = input_file.read()
                    return pl.read_avro(content, **read_kwargs)
    else:
        # fsspec AbstractFileSystem
        with filesystem.open(path, "rb", **fs_open_kwargs) as f:
            # Handle compression
            if content_encoding == ContentEncoding.IDENTITY.value:
                return pl.read_avro(f, **read_kwargs)
            else:
                input_file_init = ENCODING_TO_FILE_INIT.get(
                    content_encoding, lambda x: x
                )
                with input_file_init(f) as input_file:
                    # Read decompressed content as bytes and pass to polars
                    content = input_file.read()
                    return pl.read_avro(content, **read_kwargs)


def read_orc(
    path: str,
    *,
    filesystem: Optional[Union[AbstractFileSystem, pafs.FileSystem]] = None,
    fs_open_kwargs: Dict[str, any] = {},
    content_encoding: str = ContentEncoding.IDENTITY.value,
    **read_kwargs,
) -> pl.DataFrame:
    """
    Read an ORC file using pandas and convert to polars since polars doesn't have native ORC support.
    """
    import pandas as pd

    if not filesystem or isinstance(filesystem, pafs.FileSystem):
        path, filesystem = resolve_path_and_filesystem(path)
        with filesystem.open_input_file(path, **fs_open_kwargs) as f:
            # Handle compression
            if content_encoding == ContentEncoding.IDENTITY.value:
                pd_df = pd.read_orc(f, **read_kwargs)
                return pl.from_pandas(pd_df)
            else:
                input_file_init = ENCODING_TO_FILE_INIT.get(
                    content_encoding, lambda x: x
                )
                with input_file_init(f) as input_file:
                    # Read decompressed content and pass to pandas
                    content = input_file.read()
                    import io

                    pd_df = pd.read_orc(io.BytesIO(content), **read_kwargs)
                    return pl.from_pandas(pd_df)
    else:
        # fsspec AbstractFileSystem
        with filesystem.open(path, "rb", **fs_open_kwargs) as f:
            # Handle compression
            if content_encoding == ContentEncoding.IDENTITY.value:
                pd_df = pd.read_orc(f, **read_kwargs)
                return pl.from_pandas(pd_df)
            else:
                input_file_init = ENCODING_TO_FILE_INIT.get(
                    content_encoding, lambda x: x
                )
                with input_file_init(f) as input_file:
                    # Read decompressed content and pass to pandas
                    content = input_file.read()
                    import io

                    pd_df = pd.read_orc(io.BytesIO(content), **read_kwargs)
                    return pl.from_pandas(pd_df)


# New mapping for encoding-aware reader functions used by file_to_dataframe
CONTENT_TYPE_TO_READ_FN: Dict[str, Callable] = {
    ContentType.UNESCAPED_TSV.value: read_csv,
    ContentType.TSV.value: read_csv,
    ContentType.CSV.value: read_csv,
    ContentType.PSV.value: read_csv,
    ContentType.PARQUET.value: read_parquet,
    ContentType.FEATHER.value: read_ipc,
    ContentType.JSON.value: read_ndjson,
    ContentType.AVRO.value: read_avro,
    ContentType.ORC.value: read_orc,
}



import importlib
import logging

import deltacat.logs  # noqa: F401
from deltacat.api import (
    copy,
    get,
    list,
    put,
)
from deltacat.storage.model.transaction import (
    transaction,
    transactions,
    read_transaction,
)
from deltacat.catalog import (  # noqa: F401
    alter_namespace,
    alter_table,
    create_namespace,
    create_table,
    default_namespace,
    drop_namespace,
    drop_table,
    get_namespace,
    get_table,
    list_namespaces,
    list_tables,
    namespace_exists,
    read_table,
    refresh_table,
    rename_table,
    table_exists,
    truncate_table,
    write_to_table,
    init,
    init_local,
    is_initialized,
    clear_catalogs,
    get_catalog,
    get_catalog_properties,
    pop_catalog,
    put_catalog,
    raise_if_not_initialized,
    Catalog,
    CatalogProperties,
    TableDefinition,
)
from deltacat.compute import (
    job_client,
    local_job_client,
)
from deltacat.storage import (
    BucketingStrategy,
    BucketTransform,
    BucketTransformParameters,
    DayTransform,
    HourTransform,
    IdentityTransform,
    Dataset,
    DistributedDataset,
    Field,
    LifecycleState,
    ListResult,
    LocalDataset,
    LocalTable,
    Namespace,
    PartitionKey,
    PartitionScheme,
    Schema,
    SchemaConsistencyType,
    SortKey,
    SortOrder,
    SortScheme,
    TableProperties,
    TransactionStatus,
    Transform,
    TransformName,
    TransformParameters,
    TruncateTransform,
    TruncateTransformParameters,
    TruncateStrategy,
    UnknownTransform,
    VoidTransform,
    YearTransform,
    NullOrder,
)
from deltacat.types.media import (
    ContentEncoding,
    ContentType,
    DatasetType,
    DatastoreType,
)
from deltacat.types.tables import (
    TableWriteMode,
    TableProperty,
    TableReadOptimizationLevel,
    SchemaEvolutionMode,
    from_pandas,
    from_pyarrow,
    from_manifest_table,
    to_pyarrow,
    to_pandas,
    dataset_length,
    dataset_size,
    dataset_column_names,
    dataset_schema,
)
from deltacat.utils.url import DeltaCatUrl

write = write_to_table
read = read_table

__iceberg__ = []
if importlib.util.find_spec("pyiceberg") is not None:
    from deltacat.experimental.catalog.iceberg import (  # noqa: F401
        impl as IcebergCatalog,
    )

    __iceberg__ = [
        "IcebergCatalog",
    ]

deltacat.logs.configure_deltacat_logger(logging.getLogger(__name__))

__version__ = "2.0.0.post1"


__all__ = [
    "__version__",
    "job_client",
    "local_job_client",
    "copy",
    "get",
    "list",
    "put",
    "transaction",
    "transactions",
    "alter_table",
    "create_table",
    "drop_table",
    "refresh_table",
    "list_tables",
    "get_table",
    "truncate_table",
    "rename_table",
    "table_exists",
    "list_namespaces",
    "get_namespace",
    "namespace_exists",
    "alter_namespace",
    "create_namespace",
    "drop_namespace",
    "default_namespace",
    "write",
    "write_to_table",
    "read",
    "read_table",
    "read_transaction",
    "init",
    "init_local",
    "is_initialized",
    "clear_catalogs",
    "get_catalog",
    "get_catalog_properties",
    "pop_catalog",
    "put_catalog",
    "raise_if_not_initialized",
    "dataset_length",
    "dataset_size",
    "dataset_column_names",
    "dataset_schema",
    "from_pandas",
    "from_pyarrow",
    "from_manifest_table",
    "to_pandas",
    "to_pyarrow",
    "BucketingStrategy",
    "BucketTransform",
    "BucketTransformParameters",
    "Catalog",
    "CatalogProperties",
    "ContentType",
    "ContentEncoding",
    "Dataset",
    "DatasetType",
    "DatastoreType",
    "DayTransform",
    "DeltaCatUrl",
    "DistributedDataset",
    "Field",
    "HourTransform",
    "IdentityTransform",
    "LifecycleState",
    "ListResult",
    "LocalDataset",
    "LocalTable",
    "MonthTransform",
    "Namespace",
    "NullOrder",
    "PartitionKey",
    "PartitionScheme",
    "Schema",
    "SchemaConsistencyType",
    "SortKey",
    "SortOrder",
    "SortScheme",
    "TableDefinition",
    "TableWriteMode",
    "TableProperties",
    "TableProperty",
    "TableReadOptimizationLevel",
    "SchemaEvolutionMode",
    "TransactionStatus",
    "Transform",
    "TransformName",
    "TransformParameters",
    "TruncateTransform",
    "TruncateTransformParameters",
    "TruncateStrategy",
    "UnknownTransform",
    "VoidTransform",
    "YearTransform",
]

__all__ += __iceberg__



from typing import Protocol, Iterator, Dict, Any


class RowGenerator(Protocol):
    def generate(self) -> Dict[str, Any]:
        ...

    def generate_dataset(self, count) -> Iterator[Dict[str, Any]]:
        """Generate a dataset with a given number of records"""
        return map(lambda x: self.generate(), iter(range(count)))



from typing import List, FrozenSet, Dict

import pytest

from deltacat.experimental.storage.rivulet.metastore.delta import DeltaContext
from deltacat.experimental.storage.rivulet.metastore.sst import SSTable, SSTableRow
from deltacat.experimental.storage.rivulet.metastore.sst_interval_tree import (
    BlockIntervalTree,
    BlockGroup,
    OrderedBlockGroups,
    Block,
)
from deltacat.experimental.storage.rivulet.schema.datatype import Datatype
from deltacat.experimental.storage.rivulet import Schema


@pytest.fixture
def schema1() -> Schema:
    return Schema(
        {
            ("id", Datatype.int32()),
            ("name", Datatype.string()),
            ("age", Datatype.int32()),
        },
        "id",
    )


@pytest.fixture
def schema2() -> Schema:
    return Schema(
        {
            ("id", Datatype.int32()),
            ("address", Datatype.string()),
            ("zip", Datatype.string()),
        },
        "id",
    )


@pytest.fixture
def sst_row_list() -> List[SSTableRow]:
    return [
        SSTableRow(0, 100, "block1", 0, 1),
        SSTableRow(3, 90, "block2", 0, 1),
        SSTableRow(10, 95, "block3", 0, 1),
        SSTableRow(0, 10, "block4", 0, 1),
        SSTableRow(0, 100, "block5", 0, 1),
    ]


@pytest.fixture
def sst1(sst_row_list) -> SSTable:
    return SSTable(sst_row_list[0:3], 0, 100)


@pytest.fixture
def sst2(sst_row_list) -> SSTable:
    return SSTable(sst_row_list[3:5], 0, 100)


@pytest.fixture
def manifest_context1(schema1) -> DeltaContext:
    return DeltaContext(schema1, "manifest-001", 0)


@pytest.fixture
def manifest_context2(schema2) -> DeltaContext:
    return DeltaContext(schema2, "manifest-002", 1)


def with_field_group(
    context: DeltaContext, rows: List[SSTableRow], indexes: List[int]
) -> Dict[Schema, FrozenSet[Block]]:
    """Construct a BlockGroup dict for a singular field group"""
    schema = context.schema
    return {schema: frozenset([Block(rows[i], context) for i in indexes])}


@pytest.fixture
def expected_block_groups(
    manifest_context1, manifest_context2, sst_row_list
) -> List[BlockGroup]:
    return [
        BlockGroup(
            0,
            3,
            with_field_group(manifest_context1, sst_row_list, [0])
            | with_field_group(manifest_context2, sst_row_list, [3, 4]),
        ),
        BlockGroup(
            3,
            10,
            with_field_group(manifest_context1, sst_row_list, [0, 1])
            | with_field_group(manifest_context2, sst_row_list, [3, 4]),
        ),
        BlockGroup(
            10,
            90,
            with_field_group(manifest_context1, sst_row_list, [0, 1, 2])
            | with_field_group(manifest_context2, sst_row_list, [3, 4]),
        ),
        BlockGroup(
            90,
            95,
            with_field_group(manifest_context1, sst_row_list, [0, 1, 2])
            | with_field_group(manifest_context2, sst_row_list, [4]),
        ),
        BlockGroup(
            95,
            100,
            with_field_group(manifest_context1, sst_row_list, [0, 2])
            | with_field_group(manifest_context2, sst_row_list, [4]),
        ),
    ]


def test_build_sst(
    sst1,
    sst2,
    manifest_context1,
    manifest_context2,
    sst_row_list,
    expected_block_groups,
):
    t = BlockIntervalTree()
    t.add_sst_table(sst1, manifest_context1)
    t.add_sst_table(sst2, manifest_context2)

    block_groups = t.get_sorted_block_groups()
    expected = _build_ordered_block_groups(expected_block_groups)
    assert expected == block_groups


def test_build_sst_with_bounds(
    sst1,
    sst2,
    manifest_context1,
    manifest_context2,
    sst_row_list,
    expected_block_groups,
):
    t = BlockIntervalTree()
    t.add_sst_table(sst1, manifest_context1)
    t.add_sst_table(sst2, manifest_context2)

    block_groups_filtered = t.get_sorted_block_groups(20, 100)
    expected = _build_ordered_block_groups(expected_block_groups[2:])
    assert expected == block_groups_filtered

    block_groups_filtered = t.get_sorted_block_groups(96, 100)
    expected = _build_ordered_block_groups(expected_block_groups[4:])
    assert expected == block_groups_filtered

    block_groups_filtered = t.get_sorted_block_groups(0, 10)
    expected = _build_ordered_block_groups(expected_block_groups[0:3])
    assert expected == block_groups_filtered

    # Max key of 95 is inclusive of last range so it is included
    block_groups_filtered = t.get_sorted_block_groups(None, 95)
    expected = _build_ordered_block_groups(expected_block_groups)
    assert expected == block_groups_filtered

    block_groups_filtered = t.get_sorted_block_groups(None, 94)
    expected = _build_ordered_block_groups(expected_block_groups[0:4])
    assert expected == block_groups_filtered

    block_groups_filtered = t.get_sorted_block_groups(0, 10)
    expected = _build_ordered_block_groups(expected_block_groups[0:3])
    assert expected == block_groups_filtered

    block_groups_filtered = t.get_sorted_block_groups(0, 0)
    expected = _build_ordered_block_groups(expected_block_groups[0:1])
    assert expected == block_groups_filtered


def test_build_sst_with_non_zero_min_key_matching_global_min_key(manifest_context1):
    # Using a non-0 value since 0 evaluates to False
    min_key = 1
    max_key = 95

    sst_row = SSTableRow(min_key, max_key, "row-with-non-zero-min-key", 0, 1)
    t = BlockIntervalTree()
    t.add_sst_table(SSTable([sst_row], min_key, max_key), manifest_context1)

    block_groups_filtered = t.get_sorted_block_groups(min_key, min_key + 1)
    expected = _build_ordered_block_groups(
        [
            BlockGroup(
                min_key,
                max_key,
                {
                    manifest_context1.schema: frozenset(
                        [Block(sst_row, manifest_context1)]
                    )
                },
            )
        ]
    )
    assert expected == block_groups_filtered


def test_build_sst_invalid_bounds(
    sst1, sst2, schema1, schema2, sst_row_list, expected_block_groups
):
    t = BlockIntervalTree()

    with pytest.raises(ValueError):
        t.get_sorted_block_groups(10, 0)


def _build_ordered_block_groups(block_groups: List[BlockGroup]) -> OrderedBlockGroups:
    """
    Helper method to build OrderedBlockGroups from a sorted list of block groups

    """
    ordered_groups = []
    boundary_table = []
    for i, bg in enumerate(block_groups):
        boundary_table.append(bg.key_min)
        is_last = i == len(block_groups) - 1
        if is_last:
            bg = BlockGroup(bg.key_min, bg.key_max, bg.field_group_to_blocks, True)
            boundary_table.append(bg.key_max)
        ordered_groups.append(bg)

    return OrderedBlockGroups(
        ordered_groups[0].key_min,
        ordered_groups[-1].key_max,
        ordered_groups,
        boundary_table,
    )



from typing import Any, Dict, List, Optional, Union

from deltacat.catalog.model.catalog import get_catalog
from deltacat.catalog.model.table_definition import TableDefinition
from deltacat.storage.model.partition import (
    Partition,
    PartitionLocator,
    PartitionScheme,
)
from deltacat.storage.model.sort_key import SortScheme
from deltacat.storage.model.list_result import ListResult
from deltacat.storage.model.namespace import Namespace, NamespaceProperties
from deltacat.storage.model.schema import (
    Schema,
    SchemaUpdateOperations,
)
from deltacat.storage.model.table import TableProperties
from deltacat.storage.model.table_version import TableVersionProperties
from deltacat.storage.model.types import (
    Dataset,
    LifecycleState,
    StreamFormat,
)
from deltacat.storage.model.transaction import (
    Transaction,
    get_current_transaction,
)
from deltacat.types.media import ContentType
from deltacat.types.tables import (
    DatasetType,
    TableWriteMode,
)


# table functions
def write_to_table(
    data: Dataset,
    table: str,
    *args,
    namespace: Optional[str] = None,
    table_version: Optional[str] = None,
    mode: TableWriteMode = TableWriteMode.AUTO,
    content_type: ContentType = ContentType.PARQUET,
    transaction: Optional[Transaction] = None,
    catalog: Optional[str] = None,
    **kwargs,
) -> None:
    """Write local or distributed data to a table. Raises an error if the
    table does not exist and the table write mode is not CREATE or AUTO.

    When creating a table, all `create_table` parameters may be optionally
    specified as additional keyword arguments. When appending to, or replacing,
    an existing table, all `alter_table` parameters may be optionally specified
    as additional keyword arguments.

    Args:
        data: Local or distributed data to write to the table.
        table: Name of the table to write to.
        namespace: Optional namespace for the table. Uses default if not specified.
        table_version: Optional version of the table to write to. If specified,
            will create this version if it doesn't exist (in CREATE mode) or
            get this version if it exists (in other modes). If not specified,
            uses the latest version.
        mode: Write mode (AUTO, CREATE, APPEND, REPLACE, MERGE, DELETE).
        content_type: Content type used to write the data files. Defaults to PARQUET.
        transaction: Optional transaction to append write operations to instead of
            creating and committing a new transaction.
        **kwargs: Additional keyword arguments.
    """
    if (transaction or get_current_transaction()) and catalog:
        raise ValueError(
            "Transaction and catalog parameters are mutually exclusive. Please specify either transaction or catalog, not both."
        )
    catalog_obj = get_catalog(catalog)
    catalog_obj.impl.write_to_table(
        data,
        table,
        *args,
        namespace=namespace,
        table_version=table_version,
        mode=mode,
        content_type=content_type,
        transaction=transaction,
        inner=catalog_obj.inner,
        **kwargs,
    )


def read_table(
    table: str,
    *args,
    namespace: Optional[str] = None,
    table_version: Optional[str] = None,
    read_as: DatasetType = DatasetType.DAFT,
    partition_filter: Optional[List[Union[Partition, PartitionLocator]]] = None,
    max_parallelism: Optional[int] = None,
    columns: Optional[List[str]] = None,
    file_path_column: Optional[str] = None,
    transaction: Optional[Transaction] = None,
    catalog: Optional[str] = None,
    **kwargs,
) -> Dataset:
    """Read a table into a dataset.

    Args:
        table: Name of the table to read.
        namespace: Optional namespace of the table. Uses default if not specified.
        table_version: Optional specific version of the table to read.
        read_as: Dataset type to use for reading table files. Defaults to DatasetType.DAFT.
        partition_filter: Optional list of partitions to read from.
        max_parallelism: Optional maximum parallelism for data download. Defaults to the number of
            available CPU cores for local dataset type reads (i.e., members of DatasetType.local())
            and 100 for distributed dataset type reads (i.e., members of DatasetType.distributed()).
        columns: Optional list of columns to include in the result.
        file_path_column: Optional column name to add file paths to the result.
        transaction: Optional transaction to chain this read operation to. If provided, uncommitted
            changes from the transaction will be visible to this read operation.
        **kwargs: Additional keyword arguments.

    Returns:
        Dataset containing the table data.
    """
    if (transaction or get_current_transaction()) and catalog:
        raise ValueError(
            "Transaction and catalog parameters are mutually exclusive. Please specify either transaction or catalog, not both."
        )
    catalog_obj = get_catalog(catalog)
    return catalog_obj.impl.read_table(
        table,
        *args,
        namespace=namespace,
        table_version=table_version,
        read_as=read_as,
        partition_filter=partition_filter,
        max_parallelism=max_parallelism,
        columns=columns,
        file_path_column=file_path_column,
        transaction=transaction,
        inner=catalog_obj.inner,
        **kwargs,
    )


def alter_table(
    table: str,
    *args,
    namespace: Optional[str] = None,
    table_version: Optional[str] = None,
    lifecycle_state: Optional[LifecycleState] = None,
    schema_updates: Optional[SchemaUpdateOperations] = None,
    partition_updates: Optional[Dict[str, Any]] = None,
    sort_scheme: Optional[SortScheme] = None,
    table_description: Optional[str] = None,
    table_version_description: Optional[str] = None,
    table_properties: Optional[TableProperties] = None,
    table_version_properties: Optional[TableVersionProperties] = None,
    transaction: Optional[Transaction] = None,
    catalog: Optional[str] = None,
    **kwargs,
) -> None:
    """Alter deltacat table/table_version definition.

    Modifies various aspects of a table's metadata including lifecycle state,
    schema, partitioning, sort keys, description, and properties.

    Args:
        table: Name of the table to alter.
        namespace: Optional namespace of the table. Uses default namespace if not specified.
        table_version: Optional specific version of the table to alter. Defaults to the latest active version.
        lifecycle_state: New lifecycle state for the table.
        schema_updates: Schema updates to apply.
        partition_updates: Partition scheme updates to apply.
        sort_scheme: New sort scheme.
        table_description: New description for the table.
        table_version_description: New description for the table version. Defaults to `table_description` if not  specified.
        table_properties: New table properties.
        table_version_properties: New table version properties. Defaults to the current parent table properties if not specified.
        transaction: Optional transaction to use. If None, creates a new transaction.

    Returns:
        None

    Raises:
        TableNotFoundError: If the table does not already exist.
        TableVersionNotFoundError: If the specified table version or active table version does not exist.
    """
    if (transaction or get_current_transaction()) and catalog:
        raise ValueError(
            "Transaction and catalog parameters are mutually exclusive. Please specify either transaction or catalog, not both."
        )
    catalog_obj = get_catalog(catalog)
    catalog_obj.impl.alter_table(
        table,
        *args,
        namespace=namespace,
        table_version=table_version,
        lifecycle_state=lifecycle_state,
        schema_updates=schema_updates,
        partition_updates=partition_updates,
        sort_scheme=sort_scheme,
        table_description=table_description,
        table_version_description=table_version_description,
        table_properties=table_properties,
        table_version_properties=table_version_properties,
        transaction=transaction,
        inner=catalog_obj.inner,
        **kwargs,
    )


def create_table(
    table: str,
    *args,
    namespace: Optional[str] = None,
    table_version: Optional[str] = None,
    lifecycle_state: Optional[LifecycleState] = LifecycleState.ACTIVE,
    schema: Optional[Schema] = None,
    partition_scheme: Optional[PartitionScheme] = None,
    sort_keys: Optional[SortScheme] = None,
    table_description: Optional[str] = None,
    table_version_description: Optional[str] = None,
    table_properties: Optional[TableProperties] = None,
    table_version_properties: Optional[TableVersionProperties] = None,
    namespace_properties: Optional[NamespaceProperties] = None,
    content_types: Optional[List[ContentType]] = None,
    fail_if_exists: bool = True,
    transaction: Optional[Transaction] = None,
    catalog: Optional[str] = None,
    **kwargs,
) -> TableDefinition:
    """Create an empty table in the catalog.

    If a namespace isn't provided, the table will be created within the default deltacat namespace.
    Additionally if the provided namespace does not exist, it will be created for you.

    Args:
        table: Name of the table to create.
        namespace: Optional namespace for the table. Uses default namespace if not specified.
        version: Optional version identifier for the table.
        lifecycle_state: Lifecycle state of the new table. Defaults to ACTIVE.
        schema: Schema definition for the table.
        partition_scheme: Optional partitioning scheme for the table.
        sort_keys: Optional sort keys for the table.
        table_description: Optional description of the table.
        table_version_description: Optional description for the table version.
        table_properties: Optional properties for the table.
        table_version_properties: Optional properties for the table version. Defaults to the current parent table properties if not specified.
        namespace_properties: Optional properties for the namespace if it needs to be created.
        content_types: Optional list of allowed content types for the table.
        fail_if_exists: If True, raises an error if table already exists. If False, returns existing table.
        transaction: Optional transaction to use. If None, creates a new transaction.

    Returns:
        TableDefinition object for the created or existing table.

    Raises:
        TableAlreadyExistsError: If the table already exists and fail_if_exists is True.
        NamespaceNotFoundError: If the provided namespace does not exist.
    """
    if (transaction or get_current_transaction()) and catalog:
        raise ValueError(
            "Transaction and catalog parameters are mutually exclusive. Please specify either transaction or catalog, not both."
        )
    catalog_obj = get_catalog(catalog)
    return catalog_obj.impl.create_table(
        table,
        *args,
        namespace=namespace,
        table_version=table_version,
        lifecycle_state=lifecycle_state,
        schema=schema,
        partition_scheme=partition_scheme,
        sort_keys=sort_keys,
        table_description=table_description,
        table_version_description=table_version_description,
        table_version_properties=table_version_properties,
        table_properties=table_properties,
        namespace_properties=namespace_properties,
        content_types=content_types,
        fail_if_exists=fail_if_exists,
        transaction=transaction,
        inner=catalog_obj.inner,
        **kwargs,
    )


def drop_table(
    table: str,
    *args,
    namespace: Optional[str] = None,
    table_version: Optional[str] = None,
    purge: bool = False,
    transaction: Optional[Transaction] = None,
    catalog: Optional[str] = None,
    **kwargs,
) -> None:
    """Drop a table from the catalog and optionally purges underlying data.

    Args:
        name: Name of the table to drop.
        namespace: Optional namespace of the table. Uses default namespace if not specified.
        table_version: Optional specific version of the table to drop. Defaults to the latest active version.
        purge: If True, permanently delete the table data. If False, only remove from catalog.
        transaction: Optional transaction to use. If None, creates a new transaction.

    Returns:
        None

    Raises:
        TableNotFoundError: If the table does not exist.
        TableVersionNotFoundError: If the table version does not exist.
    """
    if (transaction or get_current_transaction()) and catalog:
        raise ValueError(
            "Transaction and catalog parameters are mutually exclusive. Please specify either transaction or catalog, not both."
        )
    catalog_obj = get_catalog(catalog)
    catalog_obj.impl.drop_table(
        table,
        *args,
        namespace=namespace,
        table_version=table_version,
        purge=purge,
        transaction=transaction,
        inner=catalog_obj.inner,
        **kwargs,
    )


def refresh_table(
    table: str,
    *args,
    namespace: Optional[str] = None,
    table_version: Optional[str] = None,
    transaction: Optional[Transaction] = None,
    catalog: Optional[str] = None,
    **kwargs,
) -> None:
    """Refresh metadata cached on the Ray cluster for the given table.

    Args:
        table: Name of the table to refresh.
        namespace: Optional namespace of the table. Uses default namespace if not specified.
        table_version: Optional specific version of the table to refresh. Defaults to the latest active version.
        transaction: Optional transaction to use. If None, creates a new transaction.

    Returns:
        None
    """
    if (transaction or get_current_transaction()) and catalog:
        raise ValueError(
            "Transaction and catalog parameters are mutually exclusive. Please specify either transaction or catalog, not both."
        )
    catalog_obj = get_catalog(catalog)
    catalog_obj.impl.refresh_table(
        table,
        *args,
        namespace=namespace,
        table_version=table_version,
        transaction=transaction,
        inner=catalog_obj.inner,
        **kwargs,
    )


def list_tables(
    *args,
    namespace: Optional[str] = None,
    table: Optional[str] = None,
    transaction: Optional[Transaction] = None,
    catalog: Optional[str] = None,
    **kwargs,
) -> ListResult[TableDefinition]:
    """List a page of table definitions.

    Args:
        namespace: Optional namespace to list tables from. Uses default namespace if not specified.
        table: Optional table to list its table versions. If not specified, lists the latest active version of each table in the namespace.
        transaction: Optional transaction to use. If None, creates a new transaction.

    Returns:
        ListResult containing TableDefinition objects for tables in the namespace.
    """
    if (transaction or get_current_transaction()) and catalog:
        raise ValueError(
            "Transaction and catalog parameters are mutually exclusive. Please specify either transaction or catalog, not both."
        )
    catalog_obj = get_catalog(catalog)
    return catalog_obj.impl.list_tables(
        *args,
        namespace=namespace,
        table=table,
        transaction=transaction,
        inner=catalog_obj.inner,
        **kwargs,
    )


def get_table(
    table: str,
    *args,
    namespace: Optional[str] = None,
    table_version: Optional[str] = None,
    stream_format: StreamFormat = StreamFormat.DELTACAT,
    transaction: Optional[Transaction] = None,
    catalog: Optional[str] = None,
    **kwargs,
) -> Optional[TableDefinition]:
    """Get table definition metadata.

    Args:
        name: Name of the table to retrieve.
        namespace: Optional namespace of the table. Uses default namespace if not specified.
        table_version: Optional specific version of the table to retrieve. Defaults to the latest active version.
        stream_format: Optional stream format to retrieve. Defaults to DELTACAT.
        transaction: Optional transaction to use. If None, creates a new transaction.

    Returns:
        Deltacat TableDefinition if the table exists, None otherwise. The table definition's table version will be
        None if the requested version is not found. The table definition's stream will be None if the requested stream
        format is not found.
    """
    if (transaction or get_current_transaction()) and catalog:
        raise ValueError(
            "Transaction and catalog parameters are mutually exclusive. Please specify either transaction or catalog, not both."
        )
    catalog_obj = get_catalog(catalog)
    return catalog_obj.impl.get_table(
        table,
        *args,
        namespace=namespace,
        table_version=table_version,
        stream_format=stream_format,
        transaction=transaction,
        inner=catalog_obj.inner,
        **kwargs,
    )


def truncate_table(
    table: str,
    *args,
    namespace: Optional[str] = None,
    table_version: Optional[str] = None,
    transaction: Optional[Transaction] = None,
    catalog: Optional[str] = None,
    **kwargs,
) -> None:
    """Truncate table data.

    Args:
        table: Name of the table to truncate.
        namespace: Optional namespace of the table. Uses default namespace if not specified.
        table_version: Optional specific version of the table to truncate. Defaults to the latest active version.
        transaction: Optional transaction to use. If None, creates a new transaction.

    Returns:
        None
    """
    if (transaction or get_current_transaction()) and catalog:
        raise ValueError(
            "Transaction and catalog parameters are mutually exclusive. Please specify either transaction or catalog, not both."
        )
    catalog_obj = get_catalog(catalog)
    catalog_obj.impl.truncate_table(
        table,
        *args,
        namespace=namespace,
        table_version=table_version,
        transaction=transaction,
        inner=catalog_obj.inner,
        **kwargs,
    )


def rename_table(
    table: str,
    new_name: str,
    *args,
    namespace: Optional[str] = None,
    transaction: Optional[Transaction] = None,
    catalog: Optional[str] = None,
    **kwargs,
) -> None:
    """Rename an existing table.

    Args:
        table: Current name of the table.
        new_name: New name for the table.
        namespace: Optional namespace of the table. Uses default namespace if not specified.
        transaction: Optional transaction to use. If None, creates a new transaction.

    Returns:
        None

    Raises:
        TableNotFoundError: If the table does not exist.
    """
    if (transaction or get_current_transaction()) and catalog:
        raise ValueError(
            "Transaction and catalog parameters are mutually exclusive. Please specify either transaction or catalog, not both."
        )
    catalog_obj = get_catalog(catalog)
    catalog_obj.impl.rename_table(
        table,
        new_name,
        *args,
        namespace=namespace,
        transaction=transaction,
        inner=catalog_obj.inner,
        **kwargs,
    )


def table_exists(
    table: str,
    *args,
    namespace: Optional[str] = None,
    table_version: Optional[str] = None,
    stream_format: StreamFormat = StreamFormat.DELTACAT,
    transaction: Optional[Transaction] = None,
    catalog: Optional[str] = None,
    **kwargs,
) -> bool:
    """Check if a table exists in the catalog.

    Args:
        table: Name of the table to check.
        namespace: Optional namespace of the table. Uses default namespace if not specified.
        table_version: Optional specific version of the table to check. Defaults to the latest active version.
        stream_format: Optional stream format to check. Defaults to DELTACAT.
        transaction: Optional transaction to use. If None, creates a new transaction.

    Returns:
        True if the table exists, False otherwise.
    """
    if (transaction or get_current_transaction()) and catalog:
        raise ValueError(
            "Transaction and catalog parameters are mutually exclusive. Please specify either transaction or catalog, not both."
        )
    catalog_obj = get_catalog(catalog)
    return catalog_obj.impl.table_exists(
        table,
        *args,
        namespace=namespace,
        table_version=table_version,
        stream_format=stream_format,
        transaction=transaction,
        inner=catalog_obj.inner,
        **kwargs,
    )


# namespace functions
def list_namespaces(
    *args,
    transaction: Optional[Transaction] = None,
    catalog: Optional[str] = None,
    **kwargs,
) -> ListResult[Namespace]:
    """List a page of table namespaces.

    Args:
        transaction: Optional transaction to use. If None, creates a new transaction.

    Returns:
        ListResult containing Namespace objects.
    """
    if (transaction or get_current_transaction()) and catalog:
        raise ValueError(
            "Transaction and catalog parameters are mutually exclusive. Please specify either transaction or catalog, not both."
        )
    catalog_obj = get_catalog(catalog)
    return catalog_obj.impl.list_namespaces(
        *args,
        transaction=transaction,
        inner=catalog_obj.inner,
        **kwargs,
    )


def get_namespace(
    namespace: str,
    *args,
    transaction: Optional[Transaction] = None,
    catalog: Optional[str] = None,
    **kwargs,
) -> Optional[Namespace]:
    """Get metadata for a specific table namespace.

    Args:
        namespace: Name of the namespace to retrieve.
        transaction: Optional transaction to use. If None, creates a new transaction.

    Returns:
        Namespace object if the namespace exists, None otherwise.
    """
    if (transaction or get_current_transaction()) and catalog:
        raise ValueError(
            "Transaction and catalog parameters are mutually exclusive. Please specify either transaction or catalog, not both."
        )
    catalog_obj = get_catalog(catalog)
    return catalog_obj.impl.get_namespace(
        namespace,
        *args,
        transaction=transaction,
        inner=catalog_obj.inner,
        **kwargs,
    )


def namespace_exists(
    namespace: str,
    *args,
    transaction: Optional[Transaction] = None,
    catalog: Optional[str] = None,
    **kwargs,
) -> bool:
    """Check if a namespace exists.

    Args:
        namespace: Name of the namespace to check.
        transaction: Optional transaction to use. If None, creates a new transaction.

    Returns:
        True if the namespace exists, False otherwise.
    """
    if (transaction or get_current_transaction()) and catalog:
        raise ValueError(
            "Transaction and catalog parameters are mutually exclusive. Please specify either transaction or catalog, not both."
        )
    catalog_obj = get_catalog(catalog)
    return catalog_obj.impl.namespace_exists(
        namespace,
        *args,
        transaction=transaction,
        inner=catalog_obj.inner,
        **kwargs,
    )


def create_namespace(
    namespace: str,
    *args,
    properties: Optional[NamespaceProperties] = None,
    transaction: Optional[Transaction] = None,
    catalog: Optional[str] = None,
    **kwargs,
) -> Namespace:
    """Create a new namespace.

    Args:
        namespace: Name of the namespace to create.
        properties: Optional properties for the namespace.
        transaction: Optional transaction to use. If None, creates a new transaction.

    Returns:
        Created Namespace object.

    Raises:
        NamespaceAlreadyExistsError: If the namespace already exists.
    """
    if (transaction or get_current_transaction()) and catalog:
        raise ValueError(
            "Transaction and catalog parameters are mutually exclusive. Please specify either transaction or catalog, not both."
        )
    catalog_obj = get_catalog(catalog)
    return catalog_obj.impl.create_namespace(
        namespace,
        *args,
        properties=properties,
        transaction=transaction,
        inner=catalog_obj.inner,
        **kwargs,
    )


def alter_namespace(
    namespace: str,
    *args,
    properties: Optional[NamespaceProperties] = None,
    new_namespace: Optional[str] = None,
    transaction: Optional[Transaction] = None,
    catalog: Optional[str] = None,
    **kwargs,
) -> None:
    """Alter a namespace definition.

    Args:
        namespace: Name of the namespace to alter.
        properties: Optional new properties for the namespace.
        new_namespace: Optional new name for the namespace.
        transaction: Optional transaction to use. If None, creates a new transaction.

    Returns:
        None
    """
    if (transaction or get_current_transaction()) and catalog:
        raise ValueError(
            "Transaction and catalog parameters are mutually exclusive. Please specify either transaction or catalog, not both."
        )
    catalog_obj = get_catalog(catalog)
    catalog_obj.impl.alter_namespace(
        namespace,
        *args,
        properties=properties,
        new_namespace=new_namespace,
        transaction=transaction,
        inner=catalog_obj.inner,
        **kwargs,
    )


def drop_namespace(
    namespace: str,
    *args,
    purge: bool = False,
    transaction: Optional[Transaction] = None,
    catalog: Optional[str] = None,
    **kwargs,
) -> None:
    """Drop a namespace and all of its tables from the catalog.

    Args:
        namespace: Name of the namespace to drop.
        purge: If True, permanently delete all table data in the namespace.
            If False, only removes the namespace from the catalog.
        transaction: Optional transaction to use. If None, creates a new transaction.

    Returns:
        None
    """
    if (transaction or get_current_transaction()) and catalog:
        raise ValueError(
            "Transaction and catalog parameters are mutually exclusive. Please specify either transaction or catalog, not both."
        )
    catalog_obj = get_catalog(catalog)
    catalog_obj.impl.drop_namespace(
        namespace,
        *args,
        purge=purge,
        transaction=transaction,
        inner=catalog_obj.inner,
        **kwargs,
    )


def default_namespace(
    *args,
    catalog: Optional[str] = None,
    **kwargs,
) -> str:
    """Return the default namespace for the catalog.

    Returns:
        Name of the default namespace.
    """
    catalog_obj = get_catalog(catalog)
    return catalog_obj.impl.default_namespace(*args, inner=catalog_obj.inner, **kwargs)



import logging
import json
import pyarrow as pa
import pyarrow.parquet
import pyarrow.feather
from typing import Callable, Dict

from deltacat.experimental.storage.rivulet.reader.query_expression import (
    QueryExpression,
)
from deltacat import logs

logger = logs.configure_deltacat_logger(logging.getLogger(__name__))


def export_parquet(dataset, file_uri: str, query: QueryExpression = QueryExpression()):
    records = dataset.scan(query).to_arrow()
    table = pa.Table.from_batches(records)
    pyarrow.parquet.write_table(table, file_uri)


def export_feather(dataset, file_uri: str, query: QueryExpression = QueryExpression()):
    records = dataset.scan(query).to_arrow()
    table = pa.Table.from_batches(records)
    pyarrow.feather.write_feather(table, file_uri)


def export_json(dataset, file_uri: str, query: QueryExpression = QueryExpression()):
    with open(file_uri, "w") as f:
        for batch in dataset.scan(query).to_pydict():
            json.dump(batch, f, indent=2)
            f.write("\n")


def export_dataset(dataset, file_uri: str, format: str = "parquet", query=None):
    """
    Export the dataset to a file.

    TODO: Make this pluggable for custom formats.

    Args:
        dataset: The dataset to export.
        file_uri: The URI to write the dataset to.
        format: The format to write the dataset in. Options are [parquet, feather, json].
        query: QueryExpression to filter the dataset before exporting.
    """
    # Supported format handlers
    export_handlers: Dict[str, Callable] = {
        "parquet": export_parquet,
        "feather": export_feather,
        "json": export_json,
    }

    if format not in export_handlers:
        raise ValueError(
            f"Unsupported format: {format}. Supported formats are {list(export_handlers.keys())}"
        )

    export_handlers[format](dataset, file_uri, query or QueryExpression())

    logger.info(f"Dataset exported to {file_uri} in {format} format.")
